"""Script to benchmark VAE architecture in isolation.

Usage:
    python -m sdfest.vae.scripts.benchmark_vae \
           --config initialization/configs/vae_models/mug.yaml \
           --device cuda
"""
import argparse
import time

import torch
import yoco

import sdfest
from sdfest.vae.sdf_vae import SDFVAE

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Benchmark script for SDFVAE.")
    config = yoco.load_config_from_args(
        parser, search_paths=[".", "~/.sdfest/", sdfest.__path__[0]]
    )
    sdf_vae = SDFVAE(
        sdf_size=64,
        latent_size=config["latent_size"],
        encoder_dict=config["encoder"],
        decoder_dict=config["decoder"],
        device=config["device"],
        tsdf=config["tsdf"],
    ).to(config["device"])
    sdf_vae.eval()

    # Forward only
    torch.set_grad_enabled(False)
    for _ in range(1000):
        torch.cuda.synchronize()
        start = time.time()

        test = sdf_vae.inference(n=1)

        torch.cuda.synchronize()
        end = time.time()
        print("forward only:", end - start)

    # Forward + Backward
    torch.set_grad_enabled(True)
    latent = torch.zeros(
        config["latent_size"], device=config["device"], requires_grad=True
    ).unsqueeze(0)
    for _ in range(1000):
        torch.cuda.synchronize()
        start = time.time()

        test = sdf_vae.decode(latent)
        loss = torch.sum(test)
        loss.backward()

        torch.cuda.synchronize()
        end = time.time()
        print("forward + backward:", end - start)



"""This module provides a deterministic low-dispersion grid on SO3."""
from typing import Tuple

import numpy as np
import healpy as hp


class SO3Grid:
    """Low-dispersion SO3 grid.

    This approach was introduced by Generating Uniform Incremental Grids on SO3 Using
    the Hopf Fibration, Yershova, 2010. We only generate the base grid (i.e., up to and
    including Section 5.2 of the paper), since we only need a fixed size set.

    Implementation roughly based on https://github.com/zhonge/cryodrgn
    """

    def __init__(self, resol: int):
        """Construct the SO3 grid.

        Args:
            resol: The resolution of the grid. Coarsest possible grid for 0.
        """
        self._resol = resol
        self._s1 = self._grid_s1(resol)
        self._s2_theta, self._s2_phi = self._grid_s2(resol)

    def num_cells(self) -> int:
        """Return the number of points in the grid."""
        return len(self._s1) * len(self._s2_theta)

    def hopf_to_index(self, psi, theta, phi):
        """Convert hopf coordinate to index.

        Args:
            phi: [0, 2pi)
            theta: [0, pi]
            psi: [0, 2pi)
        Returns:
            Grid index of closest point in grid.
        """
        s1_index = int(psi // (2 * np.pi / len(self._s1)))
        s2_index = hp.ang2pix(2 ** self._resol, theta, phi, nest=True)
        return s1_index * len(self._s2_theta) + s2_index

    def index_to_hopf(self, index: int) -> Tuple[float, float, float]:
        """Convert index to hopf coordinates.

        Psi: [0,2*pi)
        Theta: [0, pi]
        Phi: [0, 2*pi)

        Args:
            index: The index of the grid point.
        Returns:
            Tuple of psi, theta, phi.
        """
        s1_index = index // len(self._s2_theta)
        s2_index = index % len(self._s2_theta)
        psi = self._s1[s1_index]
        theta = self._s2_theta[s2_index]
        phi = self._s2_phi[s2_index]
        return psi, theta, phi

    def quat_to_index(self, quaternion: np.array) -> int:
        """Convert quaternion to index.

        Will convert quaternion to Hopf coordinates and look up closest Hopf coordinate.
        Closest means, closest in Hopf coordinates.

        Args:
            quaternion:
                Array of shape (4,), containing a normalized quaternion.
                The order of the quaternion is (x, y, z, w).
        Returns:
            The index of the closest (in Hopf coordinates) point.
        """
        hopf = SO3Grid._quat_to_hopf(quaternion)
        return self.hopf_to_index(*hopf)

    def index_to_quat(self, index: int) -> np.array:
        """Convert index to quaternion.

        Returns:
            Array of shape (4,), containing the normalized quaternion corresponding
            to the index.
        """
        hopf = self.index_to_hopf(index)
        return SO3Grid._hopf_to_quat(*hopf)

    @staticmethod
    def _quat_to_hopf(quaternion: np.array) -> Tuple[float, float, float]:
        """Convert quaternion to hopf coordinates.

        Args:
            quaternion:
                Array of shape (4,), containing a normalized quaternion.
                The order of the quaternion is (x, y, z, w).
        Returns:
            Tuple of psi, theta, phi.

            With psi, theta, phi in [0,2pi), [0,pi], [0,2pi) respectively.
        """
        x, y, z, w = quaternion
        psi = 2 * np.arctan2(x, w)
        theta = 2 * np.arctan2(np.sqrt(z ** 2 + y ** 2), np.sqrt(w ** 2 + x ** 2))
        phi = np.arctan2(z * w - x * y, y * w + x * z)

        # Note for the following correction use while instead of if, to support
        # float32, because atan2 range for float32 ([-np.float32(np.pi),
        # np.float32(np.pi)]) is larger than for float64 ([-np.pi,np.pi]).

        # Psi must be [0, 2pi) and wraps around at 4*pi, so this correction changes the
        # the half-sphere
        while psi < 0:
            psi += 2 * np.pi
        while psi >= 2 * np.pi:
            psi -= 2 * np.pi

        # Phi must be [0, 2pi) and wraps around at 2*pi, so this correction just makes
        # sure the angle is in the expected range
        while phi < 0:
            phi += 2 * np.pi
        while phi >= 2 * np.pi:
            phi -= 2 * np.pi
        return psi, theta, phi

    @staticmethod
    def _hopf_to_quat(psi, theta, phi):
        """Convert quaternion to hopf coordinates.

        Args:
            phi: [0, 2pi)
            theta: [0, pi]
            psi: [0, 2pi)
        Returns:
            Array of shape (4,), containing the normalized quaternion corresponding
            to the index.
        """
        quaternion = np.array(
            [
                np.cos(theta / 2) * np.sin(psi / 2),  # x
                np.sin(theta / 2) * np.cos(phi + psi / 2),  # y
                np.sin(theta / 2) * np.sin(phi + psi / 2),  # z
                np.cos(theta / 2) * np.cos(psi / 2),  # w
            ]
        )
        if quaternion[0] < 0:
            quaternion *= -1
        return quaternion

    @staticmethod
    def _grid_s1(resol):
        """Compute equidistant grid on 1-sphere.

        Args:
            resol: Resolution of grid.
        Returns:
            Center points of the grid cells."""
        points = 6 * 2 ** resol
        grid = np.linspace(0, 2 * np.pi, points, endpoint=False) + np.pi / points
        return grid

    @staticmethod
    def _grid_s2(resol):
        """Compute HEALpix coordinates of 2-sphere.

        Args:
            resol: Resolution of grid.
        Returns:
            Center points of the grid cells."""
        points_per_side = 2 ** resol
        points = 12 * points_per_side * points_per_side
        theta, phi = hp.pix2ang(points_per_side, np.arange(points), nest=True)
        return theta, phi



"""Experimental differentiable RGBD renderer.

This module is not optimized for performance.
"""
from typing import List, Optional, Tuple, Dict
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation
import time
from functools import partial
from collections import defaultdict


class Ray:
    """A single ray defined by its origin and direction."""

    def __init__(self, origin: np.array, direction: np.array, row: int, col: int):
        """Construct a single ray.

        Args:
            origin: The origin of the ray.
            direction: The direction of the ray. Will be normalized.
            row: The row of the pixel this ray corresponds to.
            col: The col of the pixel this ray corresponds to.
        """
        self.origin = origin
        self.direction = direction / np.linalg.norm(direction)
        self.row = row
        self.col = col

    def z_from_t(self, t: float):
        """Compute the z value in camera coordinates for a given t value.

        Note: the origin has no influence on this value.

        Args:
            t: The Euclidean distance from the origin.

        Returns:
            Z coordinate at distance t.
        """
        return t * self.direction[2]


class SDFObject:
    """An object represented as a voxelized signed distance field."""

    def __init__(self, sdf: np.array):
        """Initialize SDF based object.

        Args:
            sdf: The array containing the signed distance to the closest surface.
        """
        self._sdf = sdf
        self._grid_size = 2.0 / (sdf.shape[0] - 1)
        self._grid_size_inv = 1.0 / self._grid_size
        self._resolution = sdf.shape[0]
        # bounding box center
        self._ac = np.array([0, 0, 0, 1])
        # bounding box axes
        self._a = {}
        self._a["x"] = np.array([1, 0, 0])
        self._a["y"] = np.array([0, 1, 0])
        self._a["z"] = np.array([0, 0, 1])
        # nominal bounding box half-lengths (i.e., before scaling)
        self._h = {}
        self._h["x"] = 1
        self._h["y"] = 1
        self._h["z"] = 1

    def find_obb_ray_t(
        self, ray: Ray, o2w: np.array, scale: float
    ) -> Tuple[Optional[float], Optional[float]]:
        """Find t value for which a ray hits the oriented bounding box.

        This follows ray/box intersection algorithm described in Real-Time Rendering,
        Akenine-Möller, 2018.

        Returns:
            tuple(None, None) if there is no intersection.
            tuple(t_min, t_max) if there is an intersection. If the ray starts inside
            the obb t_min will be 0. t_max indicates the t value for which the ray
            exits the obb.
        """
        t_min = float("-inf")
        t_max = float("inf")

        # transform object
        a_c = (o2w @ self._ac)[0:3]  # rotate and translate
        a = {
            k: o2w[0:3, 0:3] @ v for k, v in self._a.items()
        }  # only rotate the axis vectors

        # ?
        p = a_c - ray.origin[0:3]

        for ax in self._a.keys():
            e = a[ax] @ p
            f = a[ax] @ ray.direction
            if (
                abs(f) > 1e-20
            ):  # ray is perpendicular to slab normal (i.e., parallel to slab)
                t_1 = (e + self._h[ax] * scale) / f
                t_2 = (e - self._h[ax] * scale) / f
                if t_1 > t_2:
                    t_1, t_2 = t_2, t_1
                t_min = max(t_min, t_1)
                t_max = min(t_max, t_2)
                if t_min > t_max or t_max < 0:
                    return None, None
            elif (
                -e > self._h[ax] * scale or -e < -self._h[ax] * scale
            ):  # if parallel, check if outside
                return None, None

        if t_min > 0:
            return t_min, t_max
        return 0, t_max

    def sphere_trace(
        self,
        ray: Ray,
        t_start: float,
        t_end: float,
        threshold: float,
        w2o: np.array,
        inv_scale: float,
    ) -> Optional[float]:
        """Find intersection of ray with SDF using sphere tracing.

        Args:
            ray: Ray in world coordinates.
            t_start: Minimum t value for ray to intersect with obb.
            t_end: Maximum t value for ray to intersect with obb.
            w2o: Matrix to transform world coordinates to object coordinates.

        Returns:
            None if no intersection is found.
            t value at the intersection otherwise.
        """
        origin = w2o @ ray.origin
        direction = w2o[0:3, 0:3] @ ray.direction
        count = 0
        scale = 1.0 / inv_scale

        t = t_start
        while t < t_end:
            point = origin[0:3] + t * direction
            count += 1
            val = self.trilinear(point, inv_scale) * scale
            # print(f"{ray.row} {ray.col} {t} {val}")
            if val < threshold * t:
                return t, count
            else:
                t += val
        return None, count

    def point_2_index(self, point: np.array) -> Tuple[int]:
        """Convert a 3D point to voxel indices.

        The point is assumed to be within [-1, 1) and the cell is the first index
        such that index * self._grid_size + 0.5 > point.
        """
        return tuple(
            max(
                0,
                min(
                    self._resolution - 2, int((x + 1.0) * (self._resolution - 1) / 2.0)
                ),
            )
            for x in point
        )

    def index_2_point(self, index: Tuple[int]) -> np.array:
        """Convert voxel index to 3D point.

        Args:
            index: The 3-tuple index to convert.

        Returns:
            The 3D point assuming a 2x2x2 sdf volume with (0,0,0) at the origin.
        """
        return np.array([i * self._grid_size - 1.0 for i in index])

    def trilinear(self, point: np.array, inv_scale: float):
        """Trilinear interpolation of the signed distance field at a local 3D point.

        Args:
            point: The 3D point in object coordinates to query.
            inv_scale: The inverted scale of the object.

        Returns:
            The distance to the closest surface, computed by trilinear interpolation.
        """
        point *= inv_scale
        base_index = self.point_2_index(point)
        n_indices = [
            tuple(i + j for i, j in zip(base_index, x))
            for x in [
                [0, 0, 0],
                [0, 0, 1],
                [0, 1, 0],
                [0, 1, 1],
                [1, 0, 0],
                [1, 0, 1],
                [1, 1, 0],
                [1, 1, 1],
            ]
        ]
        pos_0 = self.index_2_point(base_index)
        off = (point - pos_0) * self._grid_size_inv
        c00 = self._sdf[n_indices[0]] * (1 - off[0]) + self._sdf[n_indices[4]] * off[0]
        c01 = self._sdf[n_indices[1]] * (1 - off[0]) + self._sdf[n_indices[5]] * off[0]
        c10 = self._sdf[n_indices[2]] * (1 - off[0]) + self._sdf[n_indices[6]] * off[0]
        c11 = self._sdf[n_indices[3]] * (1 - off[0]) + self._sdf[n_indices[7]] * off[0]
        c0 = c00 * (1 - off[1]) + c10 * off[1]
        c1 = c01 * (1 - off[1]) + c11 * off[1]

        return c0 * (1 - off[2]) + c1 * off[2]


def generate_rays(width: int, height: int, fov_deg: float, c2w: np.array) -> List[Ray]:
    """Generate rays in world coordinates.

    The logic mostly follows scratchapixel's Generating Camera Rays chapter.

    Args:
        width: The width of the image.
        height: The height of the image.
        fov_deg: The field of the view of the camera.
        c2w: Camera-to-world transform matrix with shape (4,4).
    Returns:
        The list of rays each representing one pixel of the image.
    """
    rays = []
    aspect_ratio = width / height
    fov_rad = fov_deg * np.pi / 180
    origin = c2w @ np.array([0, 0, 0, 1])
    for row in range(0, height):
        for col in range(0, width):
            ndc_x = (col + 0.5) / width
            ndc_y = (row + 0.5) / height
            screen_x = (2 * ndc_x - 1) * np.tan(fov_rad / 2)
            screen_y = (1 - 2 * ndc_y) * np.tan(fov_rad / 2) / aspect_ratio
            pixel_pos_camera = np.array([screen_x, screen_y, -1, 1])
            pixel_pos_world = c2w @ pixel_pos_camera
            direction = pixel_pos_world[0:3] - origin[0:3]
            rays.append(Ray(origin=origin, direction=direction, row=row, col=col))

    return rays


def render_depth(
    sdf_object: SDFObject,
    width: int,
    height: int,
    fov_deg: float,
    value: str,
    threshold: float,
    p_o2w: np.array,
    q_o2w: np.array,
    inv_scale: float,
) -> np.array:
    """Render depth image of the SDF."""
    start_time = time.time()
    derivative_time = 0
    # fmt: off
    c2w = np.array([[1, 0, 0, 0],
                    [0, 1, 0, 0],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1]], dtype=np.float)
    o2w = np.array([[1, 0, 0, p_o2w[0]],
                    [0, 1, 0, p_o2w[1]],
                    [0, 0, 1, p_o2w[2]],
                    [0, 0, 0, 1]], dtype=np.float)
    o2w[0:3, 0:3] = Rotation.from_quat(q_o2w).as_matrix()
    # fmt: on

    rays = generate_rays(width, height, fov_deg, c2w)
    image = np.zeros(shape=(height, width))

    derivatives = defaultdict(partial(np.zeros_like, image))

    scale = 1.0 / inv_scale

    for ray in rays:
        t_min, t_max = sdf_object.find_obb_ray_t(ray, o2w, scale)
        # print(ray.direction, ray.row, ray.col, t_min, t_max)
        if t_min is not None:
            # print(f"{t_min} {t_max}, {ray.row} {ray.col}")
            t, c = sdf_object.sphere_trace(
                ray, t_min, t_max, threshold, np.linalg.inv(o2w), inv_scale
            )

            if t is None:
                continue
            elif value == "t":
                image[ray.row, ray.col] = t
            elif value == "z":
                image[ray.row, ray.col] = ray.z_from_t(t)
            elif value == "c":
                image[ray.row, ray.col] = c
            elif value == "d":
                image[ray.row, ray.col] = abs(ray.z_from_t(t))
                derivative_start_time = time.time()
                compute_z_derivatives(
                    derivatives, ray, t, p_o2w, q_o2w, inv_scale, sdf_object
                )
                derivative_time += time.time() - derivative_start_time
    print(
        f"Render time: {time.time()-start_time-derivative_time}"
        f"Derivative time: {derivative_time}"
    )
    return image, derivatives


def compute_z_derivatives(
    derivatives: Dict,
    ray: Ray,
    t: float,
    p_o2w: np.array,
    q_o2w: np.array,
    inv_scale: float,
    sdf_object: SDFObject,
):
    """Compute derivatives of depth value of a single pixel."""
    # see notes for derivation
    # notation: frame component var index
    x = ray.origin[0:3] + t * ray.direction
    scale = 1.0 / inv_scale

    wxx, wyx, wzx = x
    wxxo, wyxo, wzxo = p_o2w
    wxqo, wyqo, wzqo, wwqo = q_o2w
    dx, dy, dz = wxx - wxxo, wyx - wyxo, wzx - wzxo
    oxx = (
        (1 - 2 * (wyqo ** 2 + wzqo ** 2)) * dx
        + 2 * (wxqo * wyqo + wwqo * wzqo) * dy
        + 2 * (wxqo * wzqo - wwqo * wyqo) * dz
    )
    oyx = (
        2 * (wxqo * wyqo - wwqo * wzqo) * dx
        + (1 - 2 * (wxqo ** 2 + wzqo ** 2)) * dy
        + 2 * (wyqo * wzqo + wwqo * wxqo) * dz
    )
    ozx = (
        2 * (wxqo * wzqo + wwqo * wyqo) * dx
        + 2 * (wyqo * wzqo - wwqo * wxqo) * dy
        + (1 - 2 * (wxqo ** 2 + wyqo ** 2)) * dz
    )

    # retrieve cell where point terminated in
    base_index = sdf_object.point_2_index(
        np.array([oxx * inv_scale, oyx * inv_scale, ozx * inv_scale])
    )
    oxxnc, oyxnc, ozxnc = sdf_object.index_2_point(base_index)
    cxx = (-oxxnc + oxx * inv_scale) * sdf_object._grid_size_inv
    cyx = (-oyxnc + oyx * inv_scale) * sdf_object._grid_size_inv
    czx = (-ozxnc + ozx * inv_scale) * sdf_object._grid_size_inv

    n_indices = [
        tuple(i + j for i, j in zip(base_index, x))
        for x in [
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 1, 1],
            [1, 0, 0],
            [1, 0, 1],
            [1, 1, 0],
            [1, 1, 1],
        ]
    ]
    c000 = sdf_object._sdf[n_indices[0]]
    c001 = sdf_object._sdf[n_indices[1]]
    c010 = sdf_object._sdf[n_indices[2]]
    c011 = sdf_object._sdf[n_indices[3]]
    c100 = sdf_object._sdf[n_indices[4]]
    c101 = sdf_object._sdf[n_indices[5]]
    c110 = sdf_object._sdf[n_indices[6]]
    c111 = sdf_object._sdf[n_indices[7]]

    c00 = c000 * (1 - cxx) + c100 * cxx
    c01 = c001 * (1 - cxx) + c101 * cxx
    c10 = c010 * (1 - cxx) + c110 * cxx
    c11 = c011 * (1 - cxx) + c111 * cxx

    c0 = c00 * (1 - cyx) + c10 * cyx
    c1 = c01 * (1 - cyx) + c11 * cyx

    t_diff = c0 * (1 - czx) + c1 * czx

    # now compute the derivatives
    # most of what is done before, could have been cached also
    # derivatives with respect to sdf values
    if "sdf" not in derivatives:
        derivatives["sdf"] = defaultdict(derivatives.default_factory)
    f = scale * abs(ray.direction[2])
    derivatives["sdf"][n_indices[0]][ray.row, ray.col] = (
        (1 - cxx) * (1 - cyx) * (1 - czx)
    ) * f
    derivatives["sdf"][n_indices[1]][ray.row, ray.col] = (1 - cxx) * (1 - cyx) * czx * f
    derivatives["sdf"][n_indices[2]][ray.row, ray.col] = (1 - cxx) * cyx * (1 - czx) * f
    derivatives["sdf"][n_indices[3]][ray.row, ray.col] = (1 - cxx) * cyx * czx * f
    derivatives["sdf"][n_indices[4]][ray.row, ray.col] = cxx * (1 - cyx) * (1 - czx) * f
    derivatives["sdf"][n_indices[5]][ray.row, ray.col] = cxx * (1 - cyx) * czx * f
    derivatives["sdf"][n_indices[6]][ray.row, ray.col] = cxx * cyx * (1 - czx) * f
    derivatives["sdf"][n_indices[7]][ray.row, ray.col] = cxx * cyx * czx * f

    if ray.row == 20 and ray.col == 20:
        # print("Test: ", ray.direction[0], t, wxx, oxx, c000)
        print("Test: ", oxxnc, cxx)

    # derivatives with respect to object params
    s = inv_scale * sdf_object._grid_size_inv
    dcxx, dcyx, dczx = {}, {}, {}
    dcxx["x"] = (2 * (wyqo ** 2 + wzqo ** 2) - 1) * s
    dcyx["x"] = 2 * (wwqo * wzqo - wxqo * wyqo) * s
    dczx["x"] = -2 * (wxqo * wzqo + wwqo * wyqo) * s
    dcxx["y"] = -2 * (wxqo * wyqo + wwqo * wzqo) * s
    dcyx["y"] = (2 * (wxqo ** 2 + wzqo ** 2) - 1) * s
    dczx["y"] = 2 * (wwqo * wxqo - wyqo * wzqo) * s
    dcxx["z"] = 2 * (wwqo * wyqo - wxqo * wzqo) * s
    dcyx["z"] = -2 * (wyqo * wzqo + wwqo * wxqo) * s
    dczx["z"] = (2 * (wxqo ** 2 + wyqo ** 2) - 1) * s
    dcxx["qw"] = (2 * wwqo * dx + 2 * wzqo * dy - 2 * wyqo * dz - 2 * wwqo * oxx) * s
    dcyx["qw"] = (-2 * wzqo * dx + 2 * wwqo * dy + 2 * wxqo * dz - 2 * wwqo * oyx) * s
    dczx["qw"] = (2 * wyqo * dx - 2 * wxqo * dy + 2 * wwqo * dz - 2 * wwqo * ozx) * s
    dcxx["qx"] = (2 * wxqo * dx + 2 * wyqo * dy + 2 * wzqo * dz - 2 * wxqo * oxx) * s
    dcyx["qx"] = (2 * wyqo * dx - 2 * wxqo * dy + 2 * wwqo * dz - 2 * wxqo * oyx) * s
    dczx["qx"] = (2 * wzqo * dx - 2 * wwqo * dy - 2 * wxqo * dz - 2 * wxqo * ozx) * s
    dcxx["qy"] = (-2 * wyqo * dx + 2 * wxqo * dy - 2 * wwqo * dz - 2 * wyqo * oxx) * s
    dcyx["qy"] = (2 * wxqo * dx + 2 * wyqo * dy + 2 * wzqo * dz - 2 * wyqo * oyx) * s
    dczx["qy"] = (2 * wwqo * dx + 2 * wzqo * dy - 2 * wyqo * dz - 2 * wyqo * ozx) * s
    dcxx["qz"] = (-2 * wzqo * dx + 2 * wwqo * dy + 2 * wxqo * dz - 2 * wzqo * oxx) * s
    dcyx["qz"] = (-2 * wwqo * dx - 2 * wzqo * dy + 2 * wyqo * dz - 2 * wzqo * oyx) * s
    dczx["qz"] = (2 * wxqo * dx + 2 * wyqo * dy + 2 * wzqo * dz - 2 * wzqo * ozx) * s
    dcxx["s_inv"] = oxx * sdf_object._grid_size_inv
    dcyx["s_inv"] = oyx * sdf_object._grid_size_inv
    dczx["s_inv"] = ozx * sdf_object._grid_size_inv

    for k in dcxx.keys():
        dc00 = -c000 * dcxx[k] + c100 * dcxx[k]
        dc01 = -c001 * dcxx[k] + c101 * dcxx[k]
        dc10 = -c010 * dcxx[k] + c110 * dcxx[k]
        dc11 = -c011 * dcxx[k] + c111 * dcxx[k]

        dc0 = dc00 * (1 - cyx) - c00 * dcyx[k] + dc10 * cyx + c10 * dcyx[k]
        dc1 = dc01 * (1 - cyx) - c01 * dcyx[k] + dc11 * cyx + c11 * dcyx[k]

        dtdiff = dc0 * (1 - czx) - c0 * dczx[k] + dc1 * czx + c1 * dczx[k]

        derivatives[k][ray.row, ray.col] = scale * dtdiff * abs(ray.direction[2])

    # product rule for scale needed additionally
    derivatives["s_inv"][ray.row, ray.col] -= (t_diff * scale ** 2) * abs(
        ray.direction[2]
    )


def plot_derivative(derivative: np.array, cap: Optional[float] = None) -> float:
    """Plot derivative image.

    Optionally the color scale can be capped to allow multiple plots with consistent
    color scales.

    Args:
        derivative:
            HxW array containing derivative of pixel with respect to some parameter.
        cap:
            The absolute value used for the negative and positive limit of the color
            scale.
    Returns:
        The utilized maximum absolute value.
    """
    plt.figure()
    if cap is None:
        cap = abs(max(derivative.min(), derivative.max(), key=abs))
    plt.imshow(derivative, cmap="seismic", vmin=-cap, vmax=cap)
    return cap


# TODO: add click
if __name__ == "__main__":
    object = SDFObject(sdf=np.load("./data/00001.npy"))

    p_o2w = np.array([0, 0, -2.0])
    q_o2w = Rotation.from_euler("XYZ", [0, 120, 0], True).as_quat()
    inv_scale = 1.0 / 1.0
    res = 100
    image, derivatives = render_depth(
        object, res, res, 90, "d", 0.001, p_o2w, q_o2w, inv_scale=inv_scale
    )
    delta = 1e-10
    image2, _ = render_depth(
        object,
        res,
        res,
        90,
        "d",
        0.001,
        p_o2w + np.array([0, 0, 0]),
        q_o2w + np.array([0, 0, delta, 0]),
        inv_scale=inv_scale,
    )
    num_d = (image2 - image) / delta
    plt.figure()
    plt.imshow(image, cmap="gist_heat_r")
    cap = plot_derivative(derivatives["qz"])
    plot_derivative(num_d, cap)
    plt.figure()
    plt.imshow(num_d - derivatives["qz"])
    # plot_derivative(derivatives["y"])
    # plot_derivative(derivatives["z"])
    # plot_derivative(derivatives["qw"])
    # plot_derivative(derivatives["qx"])
    # plot_derivative(derivatives["qy"])
    # plot_derivative(derivatives["qz"])
    # plot_derivative(list(derivatives["sdf"].values())[500])
    plt.show()



"""Parametrized PointNet."""
import torch
import torch.nn as nn
from typing import List


class VanillaPointNet(nn.Module):
    """Parametrized PointNet without transformation layers (no T-nets).

    Generally following:
        PointNet Deep Learning on Point Sets for 3D Classification and Segmentation
        Qi, 2017
    """

    def __init__(
        self,
        in_size: int,
        mlp_out_sizes: List,
        batchnorm: bool,
        residual: bool = False,
        dense: bool = False,
    ) -> None:
        """Initialize the VanillaPointNet module.

        This module will only implements the MLP + MaxPooling part of the pointnet.

        It still requires a task specific head.

        Args:
            in_size:        dimension of the input points
            mlp_out_sizes:  output sizes of each linear layer
            batchnorm:      whether to use batchnorm or not
        """
        super().__init__()

        self._in_size = in_size
        self._mlp_out_sizes = mlp_out_sizes
        self._batchnorm = batchnorm
        self._residual = residual
        self._dense = dense

        # define layers
        self._linear_layers = torch.nn.ModuleList([])
        for i, out_size in enumerate(mlp_out_sizes):
            if i == 0:
                self._linear_layers.append(nn.Linear(self._in_size, out_size))
            else:
                if dense:
                    self._linear_layers.append(
                        nn.Linear(2 * mlp_out_sizes[i - 1], out_size)
                    )
                else:
                    self._linear_layers.append(
                        nn.Linear(mlp_out_sizes[i - 1], out_size)
                    )

        self._bn_layers = torch.nn.ModuleList([])
        if self._batchnorm:
            for out_size in mlp_out_sizes:
                self._bn_layers.append(nn.BatchNorm1d(out_size))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the module.

        Input has dimension NxMxC, where N is the batch size, M the number of points per
        set, and C the number of channels per point.

        Args:
            x: batch of point sets
        """
        set_size = x.shape[1]
        out = prev_out = x
        for i, linear_layer in enumerate(self._linear_layers):
            out = linear_layer(out)
            if self._batchnorm:
                # BN over channels across all points and sets
                pts_per_set = out.shape[1]
                out_view = out.view(-1, self._mlp_out_sizes[i])
                out = self._bn_layers[i](out_view)
                out = out.view(-1, pts_per_set, self._mlp_out_sizes[i])
            out = nn.functional.relu(out)

            if self._dense:
                out_max, _ = torch.max(out, 1, keepdim=True)
                if i != len(self._linear_layers) - 1:
                    out = torch.cat((out, out_max.expand(-1, set_size, -1)), dim=2)

            if self._residual:
                if prev_out.shape == out.shape:
                    out = prev_out + out
            prev_out = out

        # Maximum over points in same set
        out, _ = torch.max(out, 1)

        return out


class IterativePointNet(nn.Module):
    """Iterative PointNet which concatenates input.

    This is composed of 2 PointNets, where the first PointNet is applied once, the
    second PointNet a number of times, i.e.,
        out = PointNet1(in)
        for i in range(num_concat):
            out = PointNet2( concat( out, in ) )
    """

    def __init__(
        self, num_concat: int, in_size: int, mlp_out_sizes: List, batchnorm: bool
    ) -> None:
        """Initialize the IterativePointNet module.

        Args:
            num_concat:
                Number of concatenations of input and previous iteration.
                If 0 this module is the same as VanillaPointNet.
            in_size: Dimension of the input points.
            mlp_out_sizes: Output sizes of each linear layer.
            batchnorm: Whether to use batchnorm or not.
        """
        super().__init__()
        self.num_concat = num_concat
        # create 1st pointnet for taking points of channel = in_size
        self.pointnet_1 = VanillaPointNet(in_size, mlp_out_sizes, batchnorm)
        # create 2nd pointnet for taking points of channel = size of concatenated vector
        self.pointnet_2 = VanillaPointNet(
            in_size + mlp_out_sizes[-1], mlp_out_sizes, batchnorm
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Input has dimension NxMxC, where N is the batch size, M the number of points
        per set, and C the number of channels per point.

        Args:
            x: batch of point sets
        """
        # apply 1st pointnet to input
        out = self.pointnet_1(x)  # shape (batch_size, num_outputs)
        set_size = x.shape[1]
        for _ in range(self.num_concat):
            # repeat output vector across 2nd dimension
            repeated_out = out.unsqueeze(1).repeat(1, set_size, 1)
            # concatenate input vector and repeated_out
            modified_x = torch.cat((repeated_out, x), 2)
            out = self.pointnet_2(modified_x)
        return out


class GeneralizedIterativePointNet(nn.Module):
    """Generalized Iterative PointNet composed of multiple IterativePointNet instances.

    This is a sequence of iterative pointnets, where the initial input will be
    concatenated to each input, e.g.,
        out = IterativePointNet1(in)
        out = IterativePointNet2(concat(out, in))
        out = IterativePointNet3(concat(out, in))
        ...
    """

    def __init__(
        self, list_concat: list, in_size: int, list_mlp_out_sizes: list, batchnorm: bool
    ) -> None:
        """Initialize GeneralizedIterativePointnet module.

        Args:
            list_concat:
                List of concatenations for each MLP.
            in_size: Dimension of the input points.
            list_mlp_out_sizes:
                List of Output sizes of each linear layer.
                It is a List of Lists.
            batchnorm: Whether to use batchnorm or not.
        """
        super().__init__()

        init_in_size = in_size
        self.iterative_pointnet_list = torch.nn.ModuleList([])
        temp_iterative_pointnet = IterativePointNet(
            list_concat[0], in_size, list_mlp_out_sizes[0], batchnorm
        )
        self.iterative_pointnet_list.append(temp_iterative_pointnet)
        for iterative_pointnet_num in range(1, len(list_mlp_out_sizes)):
            # the input size to new MLP should be the output size of the previous MLP
            # plus previous input size
            in_size = list_mlp_out_sizes[iterative_pointnet_num - 1][-1] + init_in_size
            temp_iterative_pointnet = IterativePointNet(
                list_concat[iterative_pointnet_num],
                in_size,
                list_mlp_out_sizes[iterative_pointnet_num],
                batchnorm,
            )
            self.iterative_pointnet_list.append(temp_iterative_pointnet)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Input has dimension NxMxC, where N is the batch size, M the number of points
        per set, and C the number of channels per point.

        Args:
            x: batch of point sets
        """
        set_size = x.shape[1]
        init_x = x
        for iterative_pointnet in self.iterative_pointnet_list:
            out = iterative_pointnet(x)  # shape (batch_size, num_outputs)
            # repeat output vector across 2nd dimension
            x = out.unsqueeze(1).repeat(1, set_size, 1)
            x = torch.cat((x, init_x), 2)
        return out


if __name__ == "__main__":
    # check if dimensions are consistent across pointnets

    inp1 = torch.randn(2, 500, 3)

    pointnet = VanillaPointNet(3, [64, 64, 1024], True)
    out_p = pointnet(inp1)

    iterative_pointnet = IterativePointNet(0, 3, [64, 64, 1024], True)
    out_ip = iterative_pointnet(inp1)

    assert out_p.shape == out_ip.shape

    # check if dimension is as expected

    inp2 = torch.randn(100, 50, 2)
    iterative_pointnet2 = IterativePointNet(3, 2, [32, 64, 64, 1024], True)
    out_ip2 = iterative_pointnet2(inp2)

    assert out_ip2.shape == (100, 1024)



"""Script to play back log file and generate animation.

Usage:
  python play_log.py --log_file filename.pkl

Log files in the required format is generated by render_evaluation.py.
"""

import argparse
from datetime import datetime
import ffmpeg
import os
import pickle
from shutil import copyfile
import time

import numpy as np
import open3d as o3d
from tqdm import tqdm
import torch

from sdfest.estimation.simple_setup import SDFPipeline
from sdfest.estimation import synthetic
from sdfest.initialization import pointset_utils, quaternion_utils

reconstruction_types = ["mesh", "pointcloud", "none"]
reset = False
realtime = True
reconstruction_type = "mesh"
animation_queued = False
color = True
camera_frames = True
pause = False


def quit_program(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Quit program."""
    exit()


def reset_bounding_box(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Schedule resetting of bounding box."""
    global reset
    print("Reset bounding box.")
    reset = True
    return False


def toggle_realtime(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Toggle realtime playback."""
    global realtime
    realtime = not realtime
    print(f"Realtime: {realtime}")
    return False


def toggle_pause(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Toggle pause."""
    global pause
    pause = not pause
    print(f"Pause: {pause}")
    return False


def toggle_color(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Toggle color pointcloud."""
    global color
    color = not color
    print(f"Color: {color}")
    return False


def toggle_camera_frames(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Toggle camera frame visualization."""
    global camera_frames
    camera_frames = not camera_frames
    print(f"Camera frames: {camera_frames}")
    return False


def switch_reconstruction_type(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Switch between mesh and pointcloud reconstruction."""
    global reconstruction_type
    cur_id = reconstruction_types.index(reconstruction_type)
    cur_id = (cur_id + 1) % len(reconstruction_types)
    reconstruction_type = reconstruction_types[cur_id]
    print(f"Reconstruction: {reconstruction_type}")
    return False


def queue_animation(_: o3d.visualization.VisualizerWithKeyCallback) -> bool:
    """Switch between mesh and pointcloud reconstruction."""
    global animation_queued
    animation_queued = True
    print("Saving animation for next loop.")
    return False


def main() -> None:
    """Entry point of the evaluation program."""
    global reset
    global animation_queued
    parser = argparse.ArgumentParser(
        description="Play log file and generate animation."
    )
    parser.add_argument("--log_file", required=True)
    args = parser.parse_args()

    with open(args.log_file, "rb") as f:
        data = pickle.load(f)
        config = data["config"]
        log_entries = data["log"]

    pipeline = SDFPipeline(config)

    # precompute meshes for each timestep
    for log_entry in tqdm(log_entries):
        # print(log_entry)
        if "latent_shape" in log_entry:
            out_mesh = pipeline.generate_mesh(
                log_entry["latent_shape"], 1 / log_entry["scale_inv"], True
            )
            out_mesh.position = log_entry["position"][0].detach().cpu().numpy()
            out_mesh.orientation = log_entry["orientation"][0].detach().cpu().numpy()
            log_entry["mesh"] = out_mesh.get_transformed_o3d_geometry()
            log_entry["mesh"].paint_uniform_color([0.2, 0.4, 0.7])
            log_entry["pointcloud"] = log_entry["mesh"].sample_points_uniformly(20000)

    # visualize log entries
    cam_meshes = []
    pointclouds = []
    KEY_ESCAPE = 256
    vis = o3d.visualization.VisualizerWithKeyCallback()
    vis.register_key_callback(key=ord("A"), callback_func=reset_bounding_box)
    vis.register_key_callback(key=ord("R"), callback_func=toggle_realtime)
    vis.register_key_callback(key=ord("S"), callback_func=switch_reconstruction_type)
    vis.register_key_callback(key=ord("N"), callback_func=queue_animation)
    vis.register_key_callback(key=ord("C"), callback_func=toggle_color)
    vis.register_key_callback(key=ord("F"), callback_func=toggle_camera_frames)
    vis.register_key_callback(key=ord(" "), callback_func=toggle_pause)
    vis.register_key_callback(key=KEY_ESCAPE, callback_func=quit_program)
    vis.create_window(width=640, height=480)
    print(
        "Controls\n\ta: reset view point & bounding box\n"
        "\tr: toggle realtime\n"
        "\ts: switch reconstruction_type\n",
        "\tn: queue animation\n",
        "\tc: toggle color\n",
        "\tf: toggle camera frames\n",
        "\tspace: pause loop\n",
    )
    first = True
    while True:
        vis.clear_geometries()
        animation_folder = None
        animation_files = []
        animation_timestamps = []
        if animation_queued:
            animation_queued = False
            animation_folder = (
                f"animation_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
            )
            os.makedirs(animation_folder, exist_ok=True)
        start_time = time.time()
        for log_entry in log_entries:
            vis.poll_events()
            vis.update_renderer()
            if realtime and animation_folder is None:
                while log_entry["timestamp"] > time.time() - start_time:
                    vis.poll_events()
                    vis.update_renderer()

            if "camera_positions" in log_entry:
                cam_meshes = []
                for t_c2w, quat_c2w in zip(
                    log_entry["camera_positions"], log_entry["camera_orientations"]
                ):
                    frame_mesh = synthetic.Mesh(
                        mesh=o3d.geometry.TriangleMesh.create_coordinate_frame(
                            size=0.1, origin=[0, 0, 0]
                        ),
                        rel_scale=True,
                    )
                    frame_mesh.position = t_c2w.cpu().numpy()
                    frame_mesh.orientation = quaternion_utils.quaternion_multiply(
                        quat_c2w.cpu(), torch.tensor([1.0, 0, 0, 0])
                    ).numpy()
                    cam_meshes.append(frame_mesh.get_transformed_o3d_geometry())

            if "depth_images" in log_entry:
                pointclouds = []
                for depth_image, color_image, t_c2w, quat_c2w in zip(
                    log_entry["depth_images"],
                    log_entry["color_images"],
                    log_entry["camera_positions"],
                    log_entry["camera_orientations"],
                ):
                    points_c = pointset_utils.depth_to_pointcloud(
                        depth_image, pipeline.cam, normalize=False
                    )
                    pointcloud_torch = (
                        quaternion_utils.quaternion_apply(quat_c2w, points_c) + t_c2w
                    )
                    pointcloud_numpy = pointcloud_torch.cpu().numpy()
                    pointcloud_o3d = o3d.geometry.PointCloud(
                        o3d.utility.Vector3dVector(pointcloud_numpy)
                    )
                    if color:
                        pointcloud_colors_torch = color_image[depth_image != 0]
                        pointcloud_colors_numpy = pointcloud_colors_torch.cpu().numpy()
                        pointcloud_o3d.colors = o3d.utility.Vector3dVector(
                            pointcloud_colors_numpy
                        )
                    else:
                        pointcloud_o3d.colors = o3d.utility.Vector3dVector(
                            np.ones_like(pointcloud_numpy) * np.array([1.0, 0.2, 0.2])
                        )

                    pointclouds.append(pointcloud_o3d)

            while True:
                geometries = [] + pointclouds
                if camera_frames:
                    geometries += cam_meshes
                if "mesh" in log_entry and reconstruction_type != "none":
                    geometries.append(log_entry[reconstruction_type])
                vis.clear_geometries()
                for i, geometry in enumerate(geometries):
                    vis.add_geometry(geometry, reset_bounding_box=first or reset)
                    if i == len(geometries) - 1:
                        reset = first = False
                if not pause:
                    break
                else:
                    vis.poll_events()
                    vis.update_renderer()

            if animation_folder is not None:
                vis.poll_events()
                vis.update_renderer()
                timestamp = log_entry["timestamp"]
                filename = f"{timestamp}.png"
                vis.capture_screen_image(os.path.join(animation_folder, filename))
                animation_files.append(filename)
                animation_timestamps.append(timestamp)

        # create constant framerate video
        if animation_folder is not None:
            fps = 30
            frame_folder = os.path.join(animation_folder, "constant_framerate")
            os.makedirs(frame_folder, exist_ok=True)
            video_name = f"{animation_folder}.mp4"
            current_frame = animation_files.pop(0)
            current_time = animation_timestamps.pop(0)
            frame_number = 0
            while animation_files:
                if animation_timestamps[0] <= current_time:
                    current_frame = animation_files.pop(0)
                    animation_timestamps.pop(0)
                copyfile(
                    os.path.join(animation_folder, current_frame),
                    os.path.join(frame_folder, f"{frame_number:06d}.png"),
                )

                current_time += 1 / fps
                frame_number += 1

            ffmpeg.input(
                os.path.join(frame_folder, "*.png"), pattern_type="glob", framerate=fps
            ).output(video_name).run()


if __name__ == "__main__":
    main()



"""General functions for experiments and pytorch."""
from typing import Union

import matplotlib.pyplot as plt
import torch


def visualize_sdf_reconstruction(
    sdf: torch.Tensor, sdf_reconstruction: torch.Tensor, show: bool = False
):
    fig = plt.figure()

    plt.subplot(1, 2, 1)
    plt.subplot(1, 2, 2)

    if show:
        plt.show()

    return fig


def visualize_batch(sdfs: torch.Tensor, show: bool = False):
    fig = plt.figure()
    num_sdfs = sdfs.shape[0]

    # find nice layout
    cols = 1
    rows = 1
    while rows * cols < num_sdfs:
        if cols / rows > 4 / 3:
            rows += 1
        else:
            cols += 1

    for c in range(num_sdfs):
        plt.subplot(rows, cols, c + 1)

    if show:
        plt.show()

    return fig


def save_checkpoint(
    path: str, model: torch.nn.Module, optimizer, iteration, epoch, run_name
):
    """Save a checkpoint during training.

    Args:

    """
    torch.save(
        {
            "iteration": iteration,
            "epoch": epoch,
            "model_state_dict": model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "run_name": run_name,
        },
        path,
    )


def load_checkpoint(path, model, optimizer):
    """Load a checkpoint during training.

    Args:
    """
    print(f"Loading checkpoint at {path} ...")
    checkpoint = torch.load(path)
    model.load_state_dict(checkpoint["model_state_dict"])
    optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
    iteration = checkpoint["iteration"]
    epoch = checkpoint["epoch"]
    run_name = checkpoint["run_name"]

    print("Checkpoint loaded")

    model.train()  # training mode

    return model, optimizer, iteration, run_name, epoch


def str_to_tsdf(x: str) -> Union[bool, float]:
    """Convert string to expected values for tsdf setting.

    Args:
        x: A string containing either some representation of False or a float.
    Returns:
        False or float.
    """
    if x.lower() in ("no", "false", "f", "n", "0"):
        return False
    return float(x)


class View(torch.nn.Module):
    """Wrapper of torch's view method to use with nn.Sequential."""

    def __init__(self, shape):
        """Construct the module."""
        super(View, self).__init__()
        self.shape = shape

    def forward(self, x):
        """Reshape the tensor."""
        return x.view(*self.shape)



"""This module provides utility functions for working with SDF volumes."""
from typing import Optional

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import numpy as np
from skimage.measure import marching_cubes
import torch
import trimesh
from trimesh import Trimesh
from trimesh.visual.material import SimpleMaterial
import mesh_to_sdf as mts
import pyrender  # has to be reported after mesh_to_sdf
from pyrender.constants import RenderFlags


def mesh_to_sdf(mesh: Trimesh, cells_per_dim: int, padding: Optional[int] = 0):
    """Convert mesh to discretized signed distance field.

    The mesh will be stretched, so that its longest extend fills out the unit cube
    leaving the specified padding empty.

    Args:
        mesh: The mesh to convert.
        cells_per_dim: The number of cells along each dimension.
        padding: Number of empty space cells.
    Returns:
        The discretized signed distance field.
    """
    surface_point_method = "scan"
    sign_method = "depth"
    scaled_mesh = mts.utils.scale_to_unit_cube(mesh)
    scaled_mesh.vertices *= (cells_per_dim - 2 * padding) / cells_per_dim
    surface_point_cloud = mts.get_surface_point_cloud(
        scaled_mesh, surface_point_method, calculate_normals=sign_method == "normal"
    )
    try:
        return surface_point_cloud.get_voxels(
            cells_per_dim, check_result=True, use_depth_buffer=sign_method == "depth"
        )
    except mts.BadMeshException:
        print("Bad mesh detected. Skipping.")
        return None


def mesh_from_sdf(
    sdf_volume: np.array, level: Optional[float] = 0, complete_mesh: bool = False
) -> Trimesh:
    """Compute mesh from sdf using marching cubes algorithm.

    Args:
        sdf_volume: the SDF volume to convert, shape (M, M, M)
        level: the isosurface level to extract the mesh for
        complete_mesh:
            if True, the SDF will be padded with positive values prior to converting it
            to a mesh. This ensures a watertight mesh is created.
    Returns:
        The resulting mesh.
    """
    try:
        if complete_mesh:
            sdf_volume = np.pad(sdf_volume, pad_width=1, constant_values=1.0)
        sdf_volume.shape
        vertices, faces, normals, _ = marching_cubes(
            sdf_volume, spacing=2 / np.array(sdf_volume.shape), level=level
        )
        vertices -= 1
    except ValueError:
        return None
    return Trimesh(
        vertices,
        faces,
        vertex_normals=normals,
        visual=trimesh.visual.TextureVisuals(material=SimpleMaterial()),
    )


def plot_mesh(
    mesh: Trimesh,
    polar_angle=np.pi / 4,
    azimuth=0,
    camera_distance=2.5,
    plot_object: Optional[Axes] = None,
    transform: Optional[np.array] = None,
):
    """Render a mesh with camera pointing at its center.

    Note that in pyrender z-axis is up, x,y form the polar_angle=0 plane.

    Args:
        mesh: The mesh to render.
        polar_angle:
            Polar angle of the camera.
            For 0 the camera will look down the z-axis.
        azimuth:
            Azimuth of the camera.
            For 0, polar_anlge=pi/2 the camera will look down the x axis.
        camera_distance:
            Distance of camera to the origin.
        plot_object:
            Axis to plot the image. Will use plt if not provided.
        transform: Transform of the object. Identity by default.
    """
    if plot_object is None:
        plot_object = plt
    if transform is None:
        transform = np.eye(4, 4)[None]
    elif transform.ndim == 2:
        transform = transform[None]
    pyrender_mesh = pyrender.Mesh.from_trimesh(mesh, poses=transform, smooth=False)
    scene = pyrender.Scene()
    scene.add(pyrender_mesh)
    camera = pyrender.PerspectiveCamera(yfov=np.pi / 3.0, aspectRatio=1.0)

    # position camera on sphere centered and pointing at centroid
    camera_unit_vector = np.array(
        [
            np.sin(polar_angle) * np.cos(azimuth),
            np.sin(polar_angle) * np.sin(azimuth),
            np.cos(polar_angle),
        ]
    )
    camera_position = camera_distance * camera_unit_vector

    camera_pose = np.array(
        [
            [
                -np.sin(azimuth),
                -np.cos(azimuth) * np.cos(polar_angle),
                np.cos(azimuth) * np.sin(polar_angle),
                camera_position[0],
            ],
            [
                np.cos(azimuth),
                -np.sin(azimuth) * np.cos(polar_angle),
                np.sin(azimuth) * np.sin(polar_angle),
                camera_position[1],
            ],
            [0, np.sin(polar_angle), np.cos(polar_angle), camera_position[2]],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    scene.add(camera, pose=camera_pose)
    light = pyrender.PointLight(color=[1.0, 1.0, 1.0], intensity=45.0)
    light_pose = np.array(
        [
            [
                np.cos(azimuth) * np.cos(polar_angle),
                -np.sin(azimuth),
                np.cos(azimuth) * np.sin(polar_angle),
                camera_position[0],
            ],
            [
                np.sin(azimuth) * np.cos(polar_angle),
                np.cos(azimuth),
                np.sin(azimuth) * np.sin(polar_angle),
                camera_position[1],
            ],
            [-np.sin(polar_angle), 0, np.cos(polar_angle), camera_position[2]],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    # scene.add(light, pose=camera_pose)
    scene.add_node(pyrender.Node(light=light, matrix=light_pose))
    flags = RenderFlags.RGBA | RenderFlags.ALL_SOLID
    r = pyrender.OffscreenRenderer(1000, 1000, flags)
    color, _ = r.render(scene)
    plot_object.axis("off")
    plot_object.imshow(color, interpolation="none")


def visualize_sdf_reconstruction(
    sdf: np.array, sdf_reconstruction: np.array, show: bool = False
):
    fig = plt.figure()

    level = 1.0 / sdf.shape[-1]
    mesh = mesh_from_sdf(sdf, level=level)
    mesh_reconstruction = mesh_from_sdf(sdf_reconstruction, level=level)

    min = np.min(sdf)
    max = np.max(sdf)

    center = np.array(sdf.shape)[0] // 2

    if mesh is not None:
        plt.subplot(4, 2, 1)
        plot_mesh(mesh)
    if mesh_reconstruction is not None:
        plt.subplot(4, 2, 2)
        plot_mesh(mesh_reconstruction)

    plt.subplot(4, 2, 3)
    plt.imshow(sdf[center, :, :].T, origin="lower", vmin=min, vmax=max)
    plt.xlabel("y")
    plt.ylabel("z")
    plt.subplot(4, 2, 4)
    plt.imshow(sdf_reconstruction[center, :, :].T, origin="lower", vmin=min, vmax=max)
    plt.xlabel("y")
    plt.ylabel("z")
    plt.subplot(4, 2, 5)
    plt.imshow(sdf[:, center, :].T, origin="lower", vmin=min, vmax=max)
    plt.xlabel("x")
    plt.ylabel("z")
    plt.subplot(4, 2, 6)
    plt.imshow(sdf_reconstruction[:, center, :].T, origin="lower", vmin=min, vmax=max)
    plt.xlabel("x")
    plt.ylabel("z")
    plt.subplot(4, 2, 7)
    plt.imshow(sdf[:, :, center].T, origin="lower", vmin=min, vmax=max)
    plt.xlabel("x")
    plt.ylabel("y")
    plt.subplot(4, 2, 8)
    plt.imshow(sdf_reconstruction[:, :, center].T, origin="lower", vmin=min, vmax=max)
    plt.xlabel("x")
    plt.ylabel("y")

    if show:
        plt.show()

    return fig


def visualize_sdf_batch(sdfs: np.array, show: bool = False):
    fig = plt.figure()
    num_sdfs = sdfs.shape[0]

    level = 1.0 / sdfs.shape[-1]

    # find nice layout
    cols = 1
    rows = 1
    while rows * cols < num_sdfs:
        if cols / rows > 4 / 3:
            rows += 1
        else:
            cols += 1

    for c in range(num_sdfs):
        plt.subplot(rows, cols, c + 1)
        mesh = mesh_from_sdf(sdfs[c], level=level)
        if mesh is not None:
            plot_mesh(mesh)

    if show:
        plt.show()

    return fig


def visualize_sdf_batch_columns(sdfs: np.array, show: bool = False):
    """Visualize batch of sdfs, with one per column (mesh + cross-views)."""
    fig = plt.figure()
    num_sdfs = sdfs.shape[0]

    center = np.array(sdfs.shape)[2] // 2

    level = 1.0 / sdfs.shape[-1]

    for c in range(num_sdfs):
        min = np.min(sdfs[c])
        max = np.max(sdfs[c])

        plt.subplot(4, num_sdfs, 0 * num_sdfs + c + 1)
        mesh = mesh_from_sdf(sdfs[c], level=level)
        if mesh is not None:
            plot_mesh(mesh)
        plt.subplot(4, num_sdfs, 1 * num_sdfs + c + 1)
        plt.imshow(sdfs[c, center, :, :].T, origin="lower", vmin=min, vmax=max)
        plt.xlabel("y")
        plt.ylabel("z")
        plt.subplot(4, num_sdfs, 2 * num_sdfs + c + 1)
        plt.imshow(sdfs[c, :, center, :].T, origin="lower", vmin=min, vmax=max)
        plt.xlabel("x")
        plt.ylabel("z")
        plt.subplot(4, num_sdfs, 3 * num_sdfs + c + 1)
        plt.imshow(sdfs[c, :, :, center].T, origin="lower", vmin=min, vmax=max)
        plt.xlabel("x")
        plt.ylabel("y")

    if show:
        plt.show()

    return fig


if __name__ == "__main__":
    # test rendering
    sdf_np = np.load("./data/shapenet_mug_filtered/00000.npy")
    sdf_np2 = np.load("./data/shapenet_mug_filtered/00001.npy")
    sdf_np = np.transpose(sdf_np, (0, 2, 1))
    sdf_np2 = np.transpose(sdf_np2, (0, 2, 1))
    sdfs = np.array([sdf_np, sdf_np2])
    print(np.min(sdf_np), np.max(sdf_np))
    mesh = mesh_from_sdf(sdf_np)
    visualize_sdf_reconstruction(sdf_np, sdf_np2, show=True)
    visualize_sdf_batch(sdfs, True)
    visualize_sdf_batch_columns(sdfs, True)



"""Module for synthetic data generation."""

from abc import ABC
from typing import Optional

import open3d as o3d
from scipy.spatial.transform import Rotation
import matplotlib.pyplot as plt
import numpy as np

from sdfest.differentiable_renderer import Camera


class Object(ABC):
    """Generic positioned object representation.

    Each object has a 6-DOF pose, stored as a 3D translation vector and
    a normalized quaternion representing its orientation.
    """

    def __init__(self, position=None, orientation=None):
        """Initialize object position and orientation."""
        if position is None:
            position = np.array([0, 0, 0])
        if orientation is None:
            orientation = np.array([0, 0, 0, 1])
        self.position = position
        self.orientation = orientation


class Mesh(Object):
    """Object with associated mesh.

    This class maintains two meshes, the original mesh and the scaled mesh.

    Updating the scale will always be relative to the original mesh. I.e., two times
    setting the relative scale by 0.1 will not yield a final scale of 0.01 as it will
    always be relative to the original mesh.
    """

    def __init__(
        self,
        mesh: Optional[o3d.geometry.TriangleMesh] = None,
        path: Optional[str] = None,
        scale: float = 1,
        rel_scale: bool = False,
        center: bool = False,
        position: Optional[np.array] = None,
        orientation: Optional[np.array] = None,
    ):
        """Initialize mesh.

        Must provide either mesh or path. If path is given, mesh will be loaded from
        specified file. If mesh is given, it will be used as the original mesh.

        Args:
            mesh: The original (i.e., unscaled mesh).
            path: The path of the mesh to load.
            scale: See Mesh.update_scale.
            rel_scale: See Mesh.update_scale.
            center: whether to center the mesh upon loading.
            position: See Object.__init__.
            orientation:  See Object.__init__.
        """
        super().__init__(position=position, orientation=orientation)
        if mesh is not None and path is not None:
            raise ValueError("Only one of mesh or path can be specified")
        if mesh is not None:
            self._original_mesh = mesh
        if path is not None:
            self._original_mesh = o3d.io.read_triangle_mesh(path)
        if center:
            self._original_mesh.translate([0, 0, 0], relative=False)

        self.update_scale(scale, rel_scale)

    def load_mesh_from_file(
        self, path: str, scale: float = 1, rel_scale: bool = False
    ) -> None:
        """Load mesh from file.

        Args:
            path: Path of the obj file.
            scale: See Mesh.update_scale.
            rel_scale: See Mesh.update_scale.
        """
        self._original_mesh = o3d.io.read_triangle_mesh(path)
        self.update_scale(scale, rel_scale)

    def update_scale(self, scale: float = 1, rel_scale: bool = False) -> None:
        """Update relative or absolute scale of mesh.

        Absolute scale represents half the largest extent in x, y, or z direction.
        Relative scale represents the scale factor from original mesh.

        Args:
            scale: The desired absolute or relative scale of the object.
            rel_scale:
                If true, scale will be relative to original mesh.
                Otherwise, scale will be the resulting absolute scale.
        """
        # self._scale will always be absolute scale
        if rel_scale:
            # copy construct mesh
            self._scaled_mesh = o3d.geometry.TriangleMesh(self._original_mesh)

            original_scale = self._get_original_scale()
            self._scaled_mesh.scale(scale, [0, 0, 0])

            self._scale = original_scale * scale
        else:
            # copy construct mesh
            self._scaled_mesh = o3d.geometry.TriangleMesh(self._original_mesh)

            # scale original mesh s.t. output has the provided scale
            original_scale = self._get_original_scale()
            scale_factor = scale / original_scale
            self._scaled_mesh.scale(scale_factor, [0, 0, 0])

            self._scale = scale

    def _get_original_scale(self) -> float:
        """Compute current scale of original mesh.

        Scale is largest x/y/z extent over 2.
        """
        mins = np.amin(self._original_mesh.vertices, axis=0)
        maxs = np.amax(self._original_mesh.vertices, axis=0)
        ranges = maxs - mins
        return np.max(ranges) / 2

    def get_transformed_o3d_geometry(self) -> o3d.geometry.TriangleMesh:
        """Get o3d mesh at current pose."""
        transformed_mesh = o3d.geometry.TriangleMesh(self._scaled_mesh)
        R = Rotation.from_quat(self.orientation).as_matrix()
        transformed_mesh.rotate(R, center=np.array([0, 0, 0]))
        transformed_mesh.translate(self.position)
        transformed_mesh.compute_vertex_normals()
        return transformed_mesh


def draw_depth_geometry(obj: Object, camera: Camera):
    """Render an object given a camera."""
    # see http://www.open3d.org/docs/latest/tutorial/visualization/customized_visualization.html
    # rend = o3d.visualization.rendering.OffscreenRenderer()
    # img = rend.render_to_image()

    # Create visualizer
    vis = o3d.visualization.Visualizer()
    vis.create_window(width=camera.width, height=camera.height, visible=False)

    # Add mesh in correct position
    vis.add_geometry(obj.get_transformed_o3d_geometry(), True)

    options = vis.get_render_option()
    options.mesh_show_back_face = True

    # Set camera at fixed position (i.e., at 0,0,0, looking along z axis)
    view_control = vis.get_view_control()
    o3d_cam = camera.get_o3d_pinhole_camera_parameters()
    o3d_cam.extrinsic = np.array(
        [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
    )
    view_control.convert_from_pinhole_camera_parameters(o3d_cam, True)

    # Generate the depth image
    vis.poll_events()
    vis.update_renderer()
    depth = np.asarray(vis.capture_depth_float_buffer(do_render=True))

    return depth


if __name__ == "__main__":
    f = 320
    width = 640
    height = 480
    cam = Camera(
        width,
        height,
        fx=f,
        fy=f,
        cx=(width - 1) / 2,
        cy=(height - 1) / 2,
        pixel_center=0.5,
    )
    mesh = Mesh(path="/home/leo/datasets/shapenet_mug_filtered/00000.obj", scale=0.06)
    mesh.position = np.array([0, 0, 0.15])

    # render depth image
    depth = draw_depth_geometry(mesh, cam)

    # show depth image
    plt.imshow(depth)
    plt.show()

    # primitives
    # see: http://www.open3d.org/docs/latest/tutorial/visualization/visualization.html#Geometry-primitives



"""Application to explore and visualize VAE."""
import argparse
import datetime
import json
import os
import sys
import time
import trimesh

# need to import PySide2 before matplotlib
# see: https://matplotlib.org/3.2.1/gallery/user_interfaces/embedding_in_qt_sgskip.html # noqa
import ffmpeg
from PySide2.QtWidgets import (
    QApplication,
    QDialog,
    QListWidget,
    QListWidgetItem,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QLineEdit,
    QTabWidget,
    QWidget,
    QFileDialog,
    QLabel,
    QScrollArea,
    QSizePolicy,
    QSlider,
    QDoubleSpinBox,
)
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
from matplotlib.ticker import NullLocator
import numpy as np
from PySide2 import QtCore
import torch
import yoco

import sdfest
from sdfest.vae.sdf_vae import SDFVAE
from sdfest.vae import sdf_utils, utils


class ArgumentParserError(Exception):
    pass


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


class VAEVisualizer(QDialog):
    """User interface to explore VAE model."""

    LATENT_ABS_MAX = 4
    LATENT_TICKS = 100

    def __init__(self, parent=None):
        super(VAEVisualizer, self).__init__(parent)

        # define attributes
        self._single_sdf_input = None
        self._single_sdf_latent = None
        self._single_sdf_output = None

        self._transition_sdf_first = None
        self._transition_sdf_first_latent = None
        self._transition_sdf_first_out = None
        self._transition_sdf_second = None
        self._transition_sdf_second_latent = None
        self._transition_sdf_second_out = None
        self._transition_sdf_out = None

        self._config = None
        self._state_dict = None
        self._model = None
        self._latent_sliders = []

        self._isosurface_level = 0

        # split layout into two halfs
        # (left for loading model, right is for visualization)
        layout = QHBoxLayout()
        model_layout = QVBoxLayout()

        # Load model button
        load_model_button = QPushButton("Load model/config")
        load_model_button.clicked.connect(self.load_model)
        load_model_button.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum
        )
        model_layout.addWidget(load_model_button)

        # Confg line edit
        self._config_line_edit = QLineEdit()
        self._config_line_edit.textChanged.connect(self.config_changed)
        self._config_line_edit.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum
        )
        model_layout.addWidget(self._config_line_edit)

        # Isosurface leve spin box
        self._iso_level_spinbox = QDoubleSpinBox()
        self._iso_level_spinbox.setDecimals(2)
        self._iso_level_spinbox.setSingleStep(0.01)
        self._iso_level_spinbox.valueChanged.connect(self.update_isosurface_level)
        model_layout.addWidget(self._iso_level_spinbox)

        # Model status
        self._model_status_label = QLabel()
        model_layout.addWidget(self._model_status_label)

        # Current config
        scroll_area = QScrollArea(widgetResizable=True)
        scroll_area.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        scroll_area.setMinimumWidth(270)
        self._config_label = QLabel()
        self._config_label.setAlignment(QtCore.Qt.AlignTop)
        scroll_area.setWidget(self._config_label)

        model_layout.addWidget(scroll_area)

        # Single object explorer
        single_object_widget = QWidget()
        single_object_widget_layout = QVBoxLayout()
        single_object_widget_layout.setAlignment(QtCore.Qt.AlignTop)
        single_object_widget_bottom = QWidget()
        single_object_widget_bottom_layout = QHBoxLayout()
        single_object_save_buttons_widget = QWidget()
        single_object_save_buttons_layout = QHBoxLayout()
        self.slider_group_widget = QWidget()
        self.slider_group_widget_layout = QVBoxLayout()
        load_sdf_button = QPushButton("Load SDF")
        load_sdf_button.clicked.connect(self.load_single_sdf)
        generate_sdf_button = QPushButton("Generate SDF from prior")
        generate_sdf_button.clicked.connect(self.generate_sdf)
        save_figure_button = QPushButton("Save figure")
        save_figure_button.clicked.connect(self.save_figure)
        save_mesh_button = QPushButton("Save mesh")
        save_mesh_button.clicked.connect(self.save_mesh)
        save_sdf_button = QPushButton("Save sdf")
        save_sdf_button.clicked.connect(self.save_sdf)

        self.single_object_figure = Figure(
            dpi=85, facecolor=(1, 1, 1), edgecolor=(0, 0, 0), tight_layout=True
        )
        single_object_canvas = FigureCanvas(self.single_object_figure)
        single_object_canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.scroll_area_sliders = QScrollArea(widgetResizable=True)
        # widgetResizable indicates that the widget insided can be resized
        self.scroll_area_sliders.setFixedWidth(130)

        self.slider_group_widget.setLayout(self.slider_group_widget_layout)
        self.scroll_area_sliders.setWidget(self.slider_group_widget)
        self.scroll_area_sliders.show()

        # Single object bottom
        single_object_widget_bottom_layout.addWidget(single_object_canvas)
        single_object_widget_bottom_layout.addWidget(self.scroll_area_sliders)
        single_object_widget_bottom.setLayout(single_object_widget_bottom_layout)

        # Single object save buttons
        single_object_save_buttons_layout.addWidget(save_figure_button)
        single_object_save_buttons_layout.addWidget(save_mesh_button)
        single_object_save_buttons_layout.addWidget(save_sdf_button)
        single_object_save_buttons_widget.setLayout(single_object_save_buttons_layout)
        single_object_save_buttons_layout.setMargin(0)

        # Single object tab
        single_object_widget_layout.addWidget(load_sdf_button)
        single_object_widget_layout.addWidget(generate_sdf_button)
        single_object_widget_layout.addWidget(single_object_save_buttons_widget)
        single_object_widget_layout.addWidget(single_object_widget_bottom)
        single_object_widget.setLayout(single_object_widget_layout)

        # Transition widget
        transition_widget = QWidget()

        transition_widget_layout = QVBoxLayout()
        transition_widget_layout.setAlignment(QtCore.Qt.AlignTop)

        transition_widget_button_layout = QHBoxLayout()

        load_first_sdf_button = QPushButton("Load SDF 1")
        load_first_sdf_button.clicked.connect(lambda: self.load_transition_sdf("first"))

        load_second_sdf_button = QPushButton("Load SDF 2")
        load_second_sdf_button.clicked.connect(
            lambda: self.load_transition_sdf("second")
        )

        self.transition_figure = Figure(
            dpi=85, facecolor=(1, 1, 1), edgecolor=(0, 0, 0)
        )
        transition_canvas = FigureCanvas(self.transition_figure)
        transition_canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.transition_figure_axes = self.transition_figure.subplots(2, 3)
        self.transition_figure.delaxes(self.transition_figure_axes[0, 1])

        self.transition_slider = QSlider(QtCore.Qt.Horizontal)
        self.transition_slider.setMinimum(0)
        self.transition_slider.setMaximum(100)
        self.transition_slider.valueChanged.connect(self.update_transition_slider)

        transition_widget_button_layout.addWidget(load_first_sdf_button)
        transition_widget_button_layout.addWidget(load_second_sdf_button)
        transition_widget_layout.addLayout(transition_widget_button_layout)
        transition_widget_layout.addWidget(transition_canvas)
        transition_widget_layout.addWidget(self.transition_slider)
        transition_widget.setLayout(transition_widget_layout)

        # Animation widget
        animation_widget = QWidget()
        animation_widget_layout = QVBoxLayout()
        animation_widget_layout.setAlignment(QtCore.Qt.AlignTop)

        add_kf_from_single_view_button = QPushButton("Add from single view")
        add_kf_from_single_view_button.clicked.connect(
            lambda: self.add_kf_from_single_view()
        )

        delete_kf_button = QPushButton("Delete keyframe")
        delete_kf_button.clicked.connect(lambda: self.delete_kf())

        generate_animation_button = QPushButton("Generate animation")
        generate_animation_button.clicked.connect(lambda: self.generate_animation())

        self._keyframe_list = QListWidget()

        animation_widget.setLayout(animation_widget_layout)
        animation_widget_layout.addWidget(add_kf_from_single_view_button)
        animation_widget_layout.addWidget(delete_kf_button)
        animation_widget_layout.addWidget(self._keyframe_list)
        animation_widget_layout.addWidget(generate_animation_button)

        # Tab widget
        tab_widget = QTabWidget()
        tab_widget.addTab(single_object_widget, "Single SDF")
        tab_widget.addTab(transition_widget, "Transition")
        tab_widget.addTab(animation_widget, "Animation")

        layout.addLayout(model_layout)
        layout.addWidget(tab_widget)
        layout.setStretch(0, 0)
        layout.setStretch(1, 2)

        self.setLayout(layout)

        self.parse_config()
        self.update_model()

        # TODO make these adjustable from GUI
        self.transform = np.array(
            [
                [0, -1, 0, 0],
                [0, 0, 1, 0],
                [-1, 0, 0, 0],
                [0, 0, 0, 1],
            ]
        )
        self.azimuth = 0
        self.polar_angle = -np.pi / 4

    def reset_output(self) -> None:
        self._single_sdf_latent = None
        self._single_sdf_output = None
        self._transition_sdf_first_latent = None
        self._transition_sdf_first_out = None
        self._transition_sdf_second_latent = None
        self._transition_sdf_second_out = None

    def add_kf_from_single_view(self) -> None:
        """Add keyframe based on current latent in Single SDF tab."""
        if self._single_sdf_latent is not None:
            item = QListWidgetItem()
            item.setText(str(self._single_sdf_latent))
            item.setData(QtCore.Qt.UserRole, self._single_sdf_latent)
            self._keyframe_list.addItem(item)
        else:
            print("No valid latent currently set in Single SDF tab")

    def delete_kf(self) -> None:
        """Delete currently selected keyframe."""
        row = self._keyframe_list.currentRow()
        self._keyframe_list.takeItem(row)

    def generate_animation(self) -> None:
        """Generate animation based on keyframes."""
        fps = 30
        distance_per_frame = 0.1
        loop = True
        keyframes = []
        for i in range(self._keyframe_list.count()):
            keyframes.append(self._keyframe_list.item(i).data(QtCore.Qt.UserRole))
        if loop:
            keyframes.append(keyframes[0])

        # generate frames
        with torch.no_grad():
            fig = Figure(dpi=85, facecolor=(1, 1, 1), edgecolor=(0, 0, 0))
            now = datetime.datetime.now()
            folder = now.strftime("%Y-%m-%d_%H-%M-%S")
            os.makedirs(folder)
            frame_number = 0
            current = keyframes.pop(0).clone()
            remaining_distance = distance_per_frame
            while keyframes:
                next_target = keyframes[0]
                delta = next_target - current
                distance = torch.linalg.norm(delta)
                direction = delta / torch.linalg.norm(delta)
                if distance > distance_per_frame:
                    current += direction * distance_per_frame
                    remaining_distance = 0
                else:
                    current = keyframes.pop(0).clone()
                    remaining_distance -= distance

                if remaining_distance <= 0:
                    # generate frame
                    sdf = self._model.decode(current)
                    mesh = sdf_utils.mesh_from_sdf(
                        sdf[0, 0].cpu().numpy(),
                        self._isosurface_level,
                        complete_mesh=True,
                    )
                    if mesh is None:
                        print("No surface boundaries in SDF.")
                    else:
                        ax = fig.subplots(1, 1)
                        sdf_utils.plot_mesh(
                            mesh,
                            plot_object=ax,
                            transform=self.transform,
                            azimuth=self.azimuth,
                            polar_angle=self.polar_angle,
                        )
                        ax.xaxis.set_major_locator(NullLocator())
                        ax.yaxis.set_major_locator(NullLocator())
                        fig.savefig(
                            os.path.join(folder, f"{frame_number:05d}.png"), dpi=200
                        )
                        frame_number += 1
                        fig.clear()

        # convert frames to video
        video_name = os.path.join(f"{folder}.mp4")
        ffmpeg.input(
            os.path.join(folder, "*.png"), pattern_type="glob", framerate=fps
        ).output(video_name).run()

    def update_model(self):
        if self._config is None:
            return
        try:
            self._model = SDFVAE(
                sdf_size=64,
                latent_size=self._config["latent_size"],
                encoder_dict=self._config["encoder"],
                decoder_dict=self._config["decoder"],
                device="cpu",
                tsdf=self._config["tsdf"],
            )

            if self._state_dict is not None:
                self._model.load_state_dict(self._state_dict)
                self._model_status_label.setText(
                    "Model (with weights) successfully loaded."
                )
            else:
                self._model_status_label.setText(
                    "Model (no weights) successfully loaded."
                )
            self._model_status_label.setStyleSheet("color: green")
            self.reset_output()
        except RuntimeError:
            self._model_status_label.setText("Model config and weights do not match.")
            self._model_status_label.setStyleSheet("color: red")
            self._config_label.setText("")
            self._model = None
        except KeyError:
            self._model_status_label.setText("Model config incomplete.")
            self._model_status_label.setStyleSheet("color: red")
            self._config_label.setText("")
            self._model = None

        self.update_single()
        self.update_transition()

    def load_sdf(self):
        path, _ = QFileDialog.getOpenFileName(self, "Open file", filter="*.npy")
        if path == "":
            return None
        try:
            sdf_np = np.load(path)
            return torch.as_tensor(sdf_np).unsqueeze(0).unsqueeze(0)
        except FileNotFoundError:
            print("Error loading SDF.")
            return None

    def parse_config(self):
        try:
            arg_list = self._config_line_edit.text().split(" ")
            arg_list = [x for x in arg_list if x != ""]
            parser = ThrowingArgumentParser()
            parser.add_argument("--tsdf", type=utils.str_to_tsdf, default=False)
            parser.add_argument("--latent_size", type=lambda x: int(float(x)))
            parser.add_argument("--config", default="configs/default.yaml", nargs="+")
            self._config = yoco.load_config_from_args(parser, arg_list)
            self._config_label.setText(json.dumps(self._config, indent=4))
        except ArgumentParserError:
            self._model_status_label.setText("Error in config string.")
            self._model_status_label.setStyleSheet("color: red")
            self._config_label.setText("")
            self._model = None
            self._config = None
        except (FileNotFoundError, IsADirectoryError):
            self._model_status_label.setText("Specified config file not found.")
            self._model_status_label.setStyleSheet("color: red")
            self._config_label.setText("")
            self._model = None
            self._config = None
        except RuntimeError:
            self._model_status_label.setText("Model config and weights do not match.")
            self._model_status_label.setStyleSheet("color: red")
            self._config_label.setText("")
            self._model = None
            self._config = None

    def load_transition_sdf(self, which):
        sdf = self.load_sdf()
        if which == "first":
            self._transition_sdf_first = sdf
            self._transition_sdf_first_out = None
            self._transition_sdf_first_latent = None
        elif which == "second":
            self._transition_sdf_second = sdf
            self._transition_sdf_second_out = None
            self._transition_sdf_second_latent = None
        else:
            raise ValueError("which must be either 'first' or 'second'.")
        self.update_transition()

    def update_single(self, evaluate=True):
        # Visualize input sdf
        self.single_object_figure.clear()
        if self._single_sdf_input is not None:
            ax1, ax2 = self.single_object_figure.subplots(1, 2)
            self.render_sdf(ax1, self._single_sdf_input[0, 0].numpy())
        else:
            ax2 = self.single_object_figure.subplots(1, 1)

        if self._model is not None:
            with torch.no_grad():
                if (
                    self._single_sdf_input is not None
                    and self._single_sdf_latent is None
                ):
                    inp = self._single_sdf_input.clone()
                    self._model.prepare_input(inp)
                    self._single_sdf_latent, _, _ = self._model.encode(inp)
                    self.update_sliders()
                if self._single_sdf_latent is not None:
                    if evaluate:
                        t1 = time.time()
                        self._single_sdf_output = self._model.decode(
                            self._single_sdf_latent, enforce_tsdf=True
                        )
                        print(f"Decoding took {time.time() - t1}s")
                    t1 = time.time()
                    self.render_sdf(ax2, self._single_sdf_output[0, 0].detach().numpy())
                    print(f"Rendering took {time.time() - t1}s")

        self.single_object_figure.canvas.draw()

    def update_sliders(self):
        for current_slider in self._latent_sliders:
            self.slider_group_widget_layout.removeWidget(current_slider)
            current_slider.deleteLater()

        self._latent_sliders = []
        for latent in self._single_sdf_latent.view(-1).numpy():
            slider = QSlider(QtCore.Qt.Horizontal)
            slider.setMinimum(-self.LATENT_ABS_MAX * self.LATENT_TICKS / 2.0)
            slider.setMaximum(+self.LATENT_ABS_MAX * self.LATENT_TICKS / 2.0)
            slider.setValue(latent * self.LATENT_TICKS / 2.0)
            slider.valueChanged.connect(self.update_latent)
            self._latent_sliders.append(slider)
            self.slider_group_widget_layout.addWidget(slider)

    def update_latent(self):
        for i, current_slider in enumerate(self._latent_sliders):
            self._single_sdf_latent[0, i] = (
                current_slider.value() * 2.0 / self.LATENT_TICKS
            )
        self.update_single()

    def update_transition_slider(self):
        self.update_transition()

    def update_transition(self, update_iso_level_only=False):
        if update_iso_level_only:
            if self._transition_sdf_first_out is not None:
                self.render_sdf(
                    self.transition_figure_axes[1, 0],
                    self._transition_sdf_first_out[0, 0].detach().numpy(),
                )
            if self._transition_sdf_second_out is not None:
                self.render_sdf(
                    self.transition_figure_axes[1, 2],
                    self._transition_sdf_second_out[0, 0].detach().numpy(),
                )
            if self._transition_sdf_out is not None:
                self.render_sdf(
                    self.transition_figure_axes[1, 1],
                    self._transition_sdf_out[0, 0].detach().numpy(),
                )
            self.transition_figure.canvas.draw()
            return

        with torch.no_grad():
            if (
                self._transition_sdf_first is not None
                and self._transition_sdf_first_out is None
            ):
                print("render first sdf")
                self.render_sdf(
                    self.transition_figure_axes[0, 0],
                    self._transition_sdf_first[0, 0].detach().numpy(),
                )
                if (
                    self._model is not None
                    and self._transition_sdf_first_latent is None
                ):
                    inp = self._transition_sdf_first.clone()
                    self._model.prepare_input(inp)
                    self._transition_sdf_first_latent, _, _ = self._model.encode(inp)
                    self._transition_sdf_first_out = self._model.decode(
                        self._transition_sdf_first_latent
                    )
                    self.render_sdf(
                        self.transition_figure_axes[1, 0],
                        self._transition_sdf_first_out[0, 0].detach().numpy(),
                    )
            if (
                self._transition_sdf_second is not None
                and self._transition_sdf_second_out is None
            ):
                self.render_sdf(
                    self.transition_figure_axes[0, 2],
                    self._transition_sdf_second[0, 0].detach().numpy(),
                )
                if (
                    self._model is not None
                    and self._transition_sdf_second_latent is None
                ):
                    inp = self._transition_sdf_second.clone()
                    self._model.prepare_input(inp)
                    self._transition_sdf_second_latent, _, _ = self._model.encode(inp)
                    self._transition_sdf_second_out = self._model.decode(
                        self._transition_sdf_second_latent
                    )
                    self.render_sdf(
                        self.transition_figure_axes[1, 2],
                        self._transition_sdf_second_out[0, 0].detach().numpy(),
                    )

            if (
                self._transition_sdf_first_latent is not None
                and self._transition_sdf_second_latent is not None
            ):
                t = self.transition_slider.value() / 100.0
                interpolated_latent = (1 - t) * self._transition_sdf_first_latent + (
                    t
                ) * self._transition_sdf_second_latent
                self._transition_sdf_out = self._model.decode(interpolated_latent)
                self.render_sdf(
                    self.transition_figure_axes[1, 1],
                    self._transition_sdf_out[0, 0].detach().numpy(),
                )
        self.transition_figure.canvas.draw()

    def load_single_sdf(self):
        sdf = self.load_sdf()
        self._single_sdf_input = sdf
        self._single_sdf_latent = None
        self.update_single()

    def generate_sdf(self):
        self._single_sdf_input = None
        if self._model is not None:
            self._single_sdf_latent = self._model.sample(1) * 0.5
            self.update_single()
            self.update_sliders()
        else:
            print("Can't generate without loaded model.")

    def save_figure(self):
        """Save current single sample figure."""
        now = datetime.datetime.now()
        filename = now.strftime("%Y-%m-%d_%H-%M-%S.png")
        self.single_object_figure.savefig(filename, pad_inches=0, dpi=200)
        print(f"Saved as {filename}")

    def save_mesh(self):
        """Save current object mesh."""
        if self._single_sdf_output is None:
            print("Can't save mesh without generated SDF.")
            return
        sdf = self._single_sdf_output[0, 0].detach().numpy()
        mesh = sdf_utils.mesh_from_sdf(sdf, self._isosurface_level, complete_mesh=True)
        mesh_filename = (
            f"mesh_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')}.obj"
        )
        mesh_path = os.path.join(os.getcwd(), mesh_filename)
        with open(mesh_path, "w") as f:
            f.write(trimesh.exchange.obj.export_obj(mesh))

    def save_sdf(self):
        """Save current object sdf."""
        if self._single_sdf_output is None:
            print("Can't save SDF without generated SDF.")
            return
        sdf = self._single_sdf_output[0, 0].detach().numpy()
        sdf_filename = (
            f"sdf_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')}.npy"
        )
        sdf_path = os.path.join(os.getcwd(), sdf_filename)
        np.save(sdf_path, sdf)

    def config_changed(self, string):
        self.parse_config()
        self.update_model()

    def load_model(self):
        path, _ = QFileDialog.getOpenFileName(self, "Open file")
        if path.endswith(".yml") or path.endswith(".yaml"):
            self._config_line_edit.setText(f"--config {path}")
            self.parse_config()
            print(self._config)
            if "model" in self._config:
                self._state_dict = torch.load(self._config["model"], map_location="cpu")
            self.update_model()
        elif path.endswith(".pt"):
            self._state_dict = torch.load(path, map_location="cpu")
            self.update_model()
        else:
            return

    def render_sdf(self, ax, sdf):
        mesh = sdf_utils.mesh_from_sdf(sdf, self._isosurface_level, complete_mesh=True)
        if mesh is None:
            print("No surface boundaries in SDF.")
        else:
            sdf_utils.plot_mesh(
                mesh,
                plot_object=ax,
                transform=self.transform,
                azimuth=self.azimuth,
                polar_angle=self.polar_angle,
            )

    def update_isosurface_level(self, d):
        self._isosurface_level = d
        self.update_single(False)
        self.update_transition(True)


def main():
    """Start the UI."""
    app = QApplication(sys.argv)

    gui = VAEVisualizer()
    gui.show()

    app.exec_()


if __name__ == "__main__":
    main()



"""Differentiable rendering for signed distance fields.

By default Camera and render_depth_gpu are provided.
See the respective documentations for more information.
"""
from .sdf_renderer import Camera, render_depth_gpu

__all__ = ["Camera", "render_depth_gpu"]



import torch

if __name__ == "__main__":
    device = "cuda"
    # Conv3D speed, based on kernel size
    x = torch.rand(1,256,64,64,64, device=device)
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)   
    sizes = (1,3,5)
    for size in sizes:
        # dry run
        a = torch.nn.Conv3d(256, 1, size, padding=(size)//2).to(device)
        b = a(x)
        torch.cuda.synchronize()

        # measured run
        start.record()
        b = a(x)
        end.record()
        torch.cuda.synchronize()
        print(x.shape, "->", b.shape)
        print("Kernel Size:", size, "->", start.elapsed_time(end))

    x = torch.rand(1,3,64,64,64, device=device)
    x = x.reshape(-1, 3)
    a = torch.nn.Linear(3, 1).to(device)
    b = a(x)
    torch.cuda.synchronize()
    start.record()
    b = a(x)
    end.record()
    torch.cuda.synchronize()
    print(x.shape, "->", b.shape)
    print("Linear:", start.elapsed_time(end))
    
    # Interpolation
    x = torch.rand(1,256,32,32,32, device=device)
    a = torch.nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
    y = a(x)
    torch.cuda.synchronize()
    start.record()
    b = a(x)
    end.record()
    torch.cuda.synchronize()
    print(x.shape, "->", b.shape)
    print("Upsample tri:", start.elapsed_time(end))
    
    # Deconv
    x = torch.rand(1,256,32,32,32, device=device)
    a = torch.nn.ConvTranspose3d(256,1,3,stride=2).to(device)
    y = a(x)
    torch.cuda.synchronize()
    start.record()
    b = a(x)
    end.record()
    torch.cuda.synchronize()
    print(x.shape, "->", b.shape)
    print("Deconv:", start.elapsed_time(end))



"""Module which provides SDFDataset class."""
import math
from typing import Tuple, Optional, Iterator, TypedDict
import random

import numpy as np
from scipy.ndimage import gaussian_filter
import torch
import torchvision.transforms as T
import yoco

from sdfest.differentiable_renderer.sdf_renderer import render_depth_gpu, Camera
from sdfest.vae.sdf_vae import SDFVAE
from sdfest.initialization import pointset_utils, so3grid

# TODO support opencv / opengl conventions
# TODO support different scale conventions


class SDFVAEViewDataset(torch.utils.data.IterableDataset):
    """Dataset of SDF views generated by VAE and renderer from a random view."""
    class Config(TypedDict, total=False):
        """Configuration dictionary for SDFVAEViewDataset.

        Attributes:
            width: The width of the generated images in px.
            height: The height of the generated images in px.
            fov_deg: The horizontal fov in deg.
            z_min:
                Minimum z value (i.e., distance from camera) for the SDF.
                Note that positive z means in front of the camera, hence z_sampler
                should in most cases return positive values.
            z_max:
                Maximum z value (i.e., distance from camera) for the SDF.
            extent_mean:
                Mean extent of the SDF.
                Extent is the total side length of an SDF.
            extent_std:
                Standard deviation of the SDF scale.
            pointcloud: Whether to generate pointcloud or depth image.
            normalize_pose:
                Whether to center the augmented pointcloud at 0,0,0.
                Ignored if pointcloud=False
            orientation_repr:
                Which orientation representation is used. One of:
                    "quaternion"
                    "discretized"
            orientation_grid_resolution:
                Resolution of the orientation grid.
                Only used if orientation_repr is "discretized".
            mask_noise:
                Whether the mask should be perturbed to simulate noisy segmentation.
                If True a random, small, affine transform will be applied to the correct
                mask. The outliers will be filled with a random value sampled between
                mask_noise_min, and mask_noise_max.
            mask_noise_min:
                Minimum value to fill in for noisy mask.
                Only used if mask_noise is True.
            mask_noise_max:
                Maximum value to fill in for noisy mask.
                Only used if mask_noise is True.
            gaussian_noise_probability:
                Probability to apply gaussian noise filter on depth image.
            gaussian_noise_kernel_size:
                Size of the Gaussian kernel.
                Only used if Gaussian noise probability > 0.0.
            gausian_noise_kernel_std:
                Standard deviation of the Gaussian kernel.
                Only used if Gaussian noise probability > 0.0.
        """

        width: int
        height: int
        fov_deg: float
        z_min: float
        z_max: float
        extent_mean: float
        extent_std: float
        pointcloud: bool
        normalize_pose: Optional[bool]
        render_threshold: float
        orientation_repr: str
        orientation_grid_resolution: Optional[int]
        mask_noise: bool
        mask_noise_min: Optional[float]
        mask_noise_max: Optional[float]
        norm_noise: bool
        norm_noise_min: Optional[float]
        norm_noise_max: Optional[float]
        scale_to_unit_ball: bool
        gaussian_noise_probability: float
        gaussian_noise_kernel_size: Optional[int]
        gausian_noise_kernel_std: Optional[float]

    default_config: Config = {
        "device": "cuda",
        "width": 640,
        "height": 480,
        "fov_deg": 90,
        "render_threshold": 0.004,
        "normalize_pose": None,
        "orientation_repr": "quaternion",
        "orientation_grid_resolution": None,
        "mask_noise": False,
        "mask_noise_min": 0.1,
        "mask_noise_max": 2.0,
        "norm_noise": False,
        "norm_noise_min": -0.2,
        "norm_noise_max": 0.2,
        "scale_to_unit_ball": False,
        "gaussian_noise_probability": 0.0,
        "gaussian_noise_kernel_size": 5,
        "gaussian_noise_kernel_std": 1,
    }

    def __init__(
        self,
        config: dict,
        vae: SDFVAE,
    ) -> None:
        """Initialize the dataset.

        Args:
            config:
                Configuration dictionary of dataset. Provided dictionary will be merged
                with default_dict. See SDFVAEViewDataset.Config for supported keys.
            vae: The variational autoencoder used to create training samples.
        """
        config = yoco.load_config(config, current_dict=SDFVAEViewDataset.default_config)
        self._vae = vae
        self._vae.eval()
        self._device = next(self._vae.parameters()).device
        f = config["width"] / math.tan(config["fov_deg"] * math.pi / 180.0 / 2.0) / 2
        self._camera = Camera(
            width=config["width"],
            height=config["height"],
            fx=f,
            fy=f,
            cx=config["width"] / 2,
            cy=config["height"] / 2,
            pixel_center=0.5,
        )
        self._fov_deg = config["fov_deg"]
        self._z_min = config["z_min"]
        self._z_max = config["z_max"]
        self._z_sampler = lambda: random.uniform(self._z_min, self._z_max)
        self._extent_mean = config["extent_mean"]
        self._extent_std = config["extent_std"]
        self._scale_sampler = (
            lambda: random.gauss(self._extent_mean, self._extent_std) / 2.0
        )
        self._mask_noise = config["mask_noise"]
        self._mask_noise_min = config["mask_noise_min"]
        self._mask_noise_max = config["mask_noise_max"]
        self._mask_noise_sampler = lambda: random.uniform(
            config["mask_noise_min"], config["mask_noise_max"]
        )
        self._norm_noise = config["norm_noise"]
        self._norm_noise_min = config["norm_noise_min"]
        self._norm_noise_max = config["norm_noise_max"]
        self._norm_noise_sampler = lambda: random.uniform(
            config["norm_noise_min"], config["norm_noise_max"]
        )
        self._scale_to_unit_ball = config["scale_to_unit_ball"]
        self._pointcloud = config["pointcloud"]
        self._normalize_pose = config["normalize_pose"]
        self._render_threshold = config["render_threshold"]
        self._orientation_repr = config["orientation_repr"]
        if self._orientation_repr == "discretized":
            self._orientation_grid = so3grid.SO3Grid(
                config["orientation_grid_resolution"]
            )
        self._gaussian_noise_probability = config["gaussian_noise_probability"]
        self._create_gaussian_kernel(
            config["gaussian_noise_kernel_std"], config["gaussian_noise_kernel_size"]
        )

    def __iter__(self) -> Iterator:
        """Return SDF volume at a specific index.

        Returns:
            Infinite iterator, generating sample dictionaries.
            See SDFVAEViewDataset._generate_sample for more details about returned
            dictionaries.
        """
        # this is an infinite iterator as the sentinel False will never be returned
        while True:
            yield self._generate_valid_sample()

    def _generate_uniform_quaternion(self) -> torch.Tensor:
        """Generate a uniform quaternion.

        Following the method from K. Shoemake, Uniform Random Rotations, 1992.

        See: http://planning.cs.uiuc.edu/node198.html

        Returns:
            Uniformly distributed unit quaternion on the dataset's device.
        """
        u1, u2, u3 = random.random(), random.random(), random.random()
        return torch.tensor(
            [
                math.sqrt(1 - u1) * math.sin(2 * math.pi * u2),
                math.sqrt(1 - u1) * math.cos(2 * math.pi * u2),
                math.sqrt(u1) * math.sin(2 * math.pi * u3),
                math.sqrt(u1) * math.cos(2 * math.pi * u3),
            ]
        ).to(self._device)

    def _is_valid(self, sample: dict) -> bool:
        """Check whether a generated sample is valid.

        A valid sample contains at least one valid point in the depth image and hence
        in the pointcloud.
        """
        if sample["depth"].max() == 0:
            return False
        return True

    def _generate_valid_sample(self) -> dict:
        """Generate a single non-zero sample.

        Returns:
            See _generate_sample.
        """
        sample = self._generate_sample()
        while not self._is_valid(sample):
            sample = self._generate_sample()
            print("Warning: invalid sample, this should only happen very infrequently.")
            # if this happens often, either the SDF does not have any zero crossings
            # or the object pose is completely outside the frustum
        return sample

    def _perturb_mask(self, mask: torch.Tensor) -> torch.Tensor:
        """Perturb mask by applying small random affine transform to it.

        Args:
            mask: The mask to perturb.
        Returns:
            The perturbed mask. Same shape as mask.
        """
        affine_transfomer = T.RandomAffine(
            degrees=(0, 1), translate=(0.00, 0.01), scale=(0.999, 1.001)
        )
        return affine_transfomer(mask.unsqueeze(0))[0]

    def _generate_sample(self) -> Tuple:
        """Generate a single sample. Possibly (albeit very unlikely) zero / empty.

        Return:
            Sample containing the following items:
                "depth"
                "pointset"
                "latent_shape"
                "position"
                "orientation"
                "quaternion"
                "scale"
        """
        sample = {}

        latent = self._vae.sample()  # will be 1xlatent_size (batch size 1 here)
        with torch.no_grad():
            sdf = self._vae.decode(latent)
        # generate x, y, z s.t. center is inside frustum
        z = self._z_sampler()
        x_pix = random.uniform(-self._camera.width / 2, self._camera.height / 2)
        x = x_pix / self._camera.fx * z
        y_pix = random.uniform(-self._camera.height / 2, self._camera.height / 2)
        y = y_pix / self._camera.fy * z
        position = torch.tensor([x, y, -z]).to(self._device)
        quaternion = self._generate_uniform_quaternion()
        scale = torch.tensor(self._scale_sampler()).to(self._device)
        inv_scale = 1.0 / scale
        orientation = self._quat_to_orientation_repr(quaternion)

        depth = render_depth_gpu(
            sdf[0, 0],
            position,
            quaternion,
            inv_scale,
            threshold=self._render_threshold,
            camera=self._camera,
        )

        exact_mask = depth != 0
        if self._mask_noise:
            final_mask = self._perturb_mask(exact_mask)
            depth[~exact_mask] = self._mask_noise_sampler()
        else:
            final_mask = exact_mask

        if self._gaussian_noise_probability > 0.0:
            if random.random() < self._gaussian_noise_probability:
                invalid_depth_mask = depth == 0
                depth[invalid_depth_mask] = torch.nan
                depth_filtered = torch.nn.functional.conv2d(
                    depth[None, None], self._gaussian_kernel, padding="same"
                )[0, 0]
                # nan might become inf be preserved
                # https://github.com/pytorch/pytorch/issues/12484
                mask = torch.logical_or(depth_filtered.isnan(), depth_filtered.isinf())
                depth[~mask] = depth_filtered[~mask]
                depth[depth.isnan()] = 0.0

        depth[~final_mask] = 0

        if self._pointcloud:
            pointset = pointset_utils.depth_to_pointcloud(
                depth, self._camera, convention="opengl"
            )

            if self._normalize_pose:
                pointset, centroid = pointset_utils.normalize_points(pointset)
                position -= centroid  # adjust target

                if self._norm_noise:
                    noise = position.new_tensor(
                        [
                            self._norm_noise_sampler(),
                            self._norm_noise_sampler(),
                            self._norm_noise_sampler(),
                        ]
                    )
                    position += noise
                    pointset += noise

                if self._scale_to_unit_ball:
                    max_distance = torch.max(torch.linalg.norm(pointset))
                    pointset /= max_distance
                    scale /= max_distance

            sample["pointset"] = pointset

        sample["depth"] = depth
        sample["latent_shape"] = latent.squeeze()
        sample["position"] = position
        sample["orientation"] = orientation
        sample["quaternion"] = quaternion
        sample["scale"] = scale

        return sample

        # TODO: add augmentation

    def _quat_to_orientation_repr(self, quaternion: torch.Tensor) -> torch.Tensor:
        """Convert quaternion to selected orientation representation.

        Args:
            quaternion:
                The quaternion to convert, scalar-last, shape (4,).
        Returns:
            The same orientation as represented by the quaternion in the chosen
            orientation representation.
        """
        if self._orientation_repr == "quaternion":
            return quaternion
        elif self._orientation_repr == "discretized":
            index = self._orientation_grid.quat_to_index(quaternion.cpu().numpy())
            return torch.tensor(
                index,
                device=self._device,
                dtype=torch.long,
            )
        else:
            raise NotImplementedError(
                f"Orientation representation {self._orientation_repr} is not supported."
            )

    def _create_gaussian_kernel(self, std: float, kernel_size: int) -> None:
        """Create and set Gaussian noise kernel used for smoothing the depth image."""
        if kernel_size % 2 != 1:
            raise ValueError("Kernel size should be odd.")
        impulse = np.zeros((kernel_size, kernel_size))
        impulse[kernel_size // 2, kernel_size // 2] = 1
        kernel = gaussian_filter(impulse, std)
        self._gaussian_kernel = torch.Tensor(kernel[None, None]).to(self._device)



"""Module for utility function related to NOCS dataset.

This module contains functions to find similarity transform from NOCS maps and
evaluation function for typical metrics on the NOCS datasets.

Aligning code by Srinath Sridhar:
    https://raw.githubusercontent.com/hughw19/NOCS_CVPR2019/master/aligning.py

Evaluation code by ... TODO
"""
import numpy as np


def estimate_similarity_transform(
    source: np.ndarray, target: np.ndarray, verbose: bool = False
) -> tuple:
    """Estimate similarity transform from source to target from point correspondences.

    Source and target are pairwise correponding pointsets, i.e., they include same
    number of points and the first point of source corresponds to the first point of
    target. RANSAC is used for outlier-robust estimation.

    A similarity transform is estimated (i.e., isotropic scale, rotation and
    translation) that transforms source points onto the target points.

    Note that the returned values fulfill the following equations
        transform @ source_points = scale * rotation_matrix @ source_points + position
    when ignoring homogeneous coordinate for left-hand side.

    Args:
        source: Source points that will be transformed, shape (N,3).
        target: Target points to which source will be aligned to, shape (N,3).
        verbose: If true additional information will be printed.
    Returns:
        position (np.ndarray): Translation to translate source to target, shape (3,).
        rotation_matrix (np.ndarray): Rotation to rotate source to target, shape (3,3).
        scale (float):
            Scaling factor along each axis, to scale source to target.
        transform (np.ndarray): Homogeneous transformation matrix, shape (4,4).
    """
    if len(source) < 5 or len(target) < 5:
        print("Pose estimation failed. Not enough point correspondences: ", len(source))
        return None, None, None, None

    # make points homogeneous
    source_hom = np.transpose(np.hstack([source, np.ones([source.shape[0], 1])]))  # 4,N
    target_hom = np.transpose(np.hstack([target, np.ones([source.shape[0], 1])]))  # 4,N

    # Auto-parameter selection based on source-target heuristics
    target_norm = np.mean(np.linalg.norm(target, axis=1))  # mean distance from origin
    source_norm = np.mean(np.linalg.norm(source, axis=1))
    ratio_ts = target_norm / source_norm
    ratio_st = source_norm / target_norm
    pass_t = ratio_st if (ratio_st > ratio_ts) else ratio_ts
    pass_t *= 0.01  # tighter bound
    stop_t = pass_t / 100
    n_iter = 100
    if verbose:
        print("Pass threshold: ", pass_t)
        print("Stop threshold: ", stop_t)
        print("Number of iterations: ", n_iter)

    source_inliers_hom, target_inliers_hom, best_inlier_ratio = _get_ransac_inliers(
        source_hom,
        target_hom,
        max_iterations=n_iter,
        pass_threshold=pass_t,
        stop_threshold=stop_t,
    )

    if best_inlier_ratio < 0.1:
        print("Pose estimation failed. Small inlier ratio: ", best_inlier_ratio)
        return None, None, None, None

    scales, rotation_matrix, position, out_transform = _estimate_similarity_umeyama(
        source_inliers_hom, target_inliers_hom
    )
    scale = scales[0]

    if verbose:
        print("BestInlierRatio:", best_inlier_ratio)
        print("Rotation:\n", rotation_matrix)
        print("Position:\n", position)
        print("Scales:", scales)

    return position, rotation_matrix, scale, out_transform


def _get_ransac_inliers(
    source_hom: np.ndarray,
    target_hom: np.ndarray,
    max_iterations: int = 100,
    pass_threshold: float = 200,
    stop_threshold: float = 1,
) -> tuple:
    """Apply RANSAC and return set of inliers.

    Args:
        source_hom: Homogeneous coordinates of source points, shape (4,N).
        target_hom: Homogeneous coordinates of target points, shape (4,N).
        max_iterations: Maximum number of RANSAC iterations.
        pass_threshold: Threshold at which a point correspondence is considered good.
        stop_threshold: If residual is below this threshold, RANSAC will stop early.
    Returns:
        source_inliers (np.ndarray):
            Homogeneous coordinates of inlier source points, shape (4,M).
        target_inliers (np.ndarray):
            Homogeneous coordinates of inlier target points, shape (4,M).
        inlier_ratio (float): Ratio of inliers and outliers.
    """
    best_residual = 1e10
    best_inlier_ratio = 0
    best_inlier_idx = np.arange(source_hom.shape[1])
    for _ in range(0, max_iterations):
        # Pick 5 random (but corresponding) points from source and target without repl.
        rand_idx = np.random.choice(
            source_hom.shape[1], size=5, replace=False
        )
        # rand_idx = np.random.randint(source_hom.shape[1], size=10)
        _, _, _, out_transform = _estimate_similarity_umeyama(
            source_hom[:, rand_idx], target_hom[:, rand_idx]
        )
        residual, inlier_ratio, inlier_idx = _evaluate_model(
            out_transform, source_hom, target_hom, pass_threshold
        )
        if residual < best_residual:
            best_residual = residual
            best_inlier_ratio = inlier_ratio
            best_inlier_idx = inlier_idx
        if best_residual < stop_threshold:
            break

        # print('Iteration: ', i)
        # print('Residual: ', Residual)
        # print('Inlier ratio: ', InlierRatio)

    return (
        source_hom[:, best_inlier_idx],
        target_hom[:, best_inlier_idx],
        best_inlier_ratio,
    )


def _evaluate_model(
    out_transform: np.ndarray,
    source_hom: np.ndarray,
    target_hom: np.ndarray,
    pass_threshold: float,
) -> tuple:
    """Evaluate transformation from source to target points.

    Args:
        out_transform:
            Transformation which will be applied to source points, shape (4,4).
        source_hom: Homogeneous coordinates of source points, shape (4,N).
        target_hom: Homogeneous coordinates of target points, shape (4,N).
        pass_threshold: Threshold at which a point correspondence is considered good.
    Returns:
        residual (float): The mean error between transformed source and target.
        inlier_ratio (float):
            Ratio between inliers and number of correspondences (i.e., N).
        inlier_idx (np.ndarray): Array containing the indices of inliers, shape (M,).
    """
    diff = target_hom - np.matmul(out_transform, source_hom)
    residual_vec = np.linalg.norm(diff[:3, :], axis=0)  # shape (N,)
    residual = np.linalg.norm(residual_vec)
    inlier_idx = np.where(residual_vec < pass_threshold)
    n_inliers = np.count_nonzero(inlier_idx)
    inlier_ratio = n_inliers / source_hom.shape[1]
    return residual, inlier_ratio, inlier_idx[0]


def _estimate_similarity_umeyama(
    source_hom: np.ndarray, target_hom: np.ndarray
) -> tuple:
    """Calculate similarity transform from 3D point correspondences.

    A similarity transform is calculated (i.e., isotropic scale, rotation and
    translation) that transforms source points such that squared Euclidean distance
    between transformed source points and target points is minimized.

    Original algorithm from Least-squares estimation of transformation parameters
    between two point patterns, Umeyama, 1991.
    http://web.stanford.edu/class/cs273/refs/umeyama.pdf

    The returned scale, rotation, translation, describe the same transformation as the
    homogeneous transformation matrix as (np notation):
        scale * rotation @ point + translation <=> transformation @ point_hom.

    Args:
        source_hom: Homogeneous coordinates of 5 source points, shape (4,M).
        target_hom: Homogeneous coordinates of 5 target points, shape (4,M).
    Returns:
        scales (np.ndarray):
            Scaling factors along each axis, to scale source to target, shape (3,).
            This will always be three times the same value, since similarity transforms
            only include isotropic scaling.
        rotation (np.ndarray): Rotation to rotate source to target, shape (3,).
        translation (np.ndarray): Translation to translate source to target, shape (3,).
        transform (np.ndarray): Homogeneous transformation matrix, shape (4,4).
    """
    source_centroid = np.mean(source_hom[:3, :], axis=1)  # shape (3,)
    target_centroid = np.mean(target_hom[:3, :], axis=1)  # shape (3,)
    n_points = source_hom.shape[1]

    centered_source = source_hom[:3, :] - source_centroid[:, None]  # shape (3, N)
    centered_target = target_hom[:3, :] - target_centroid[:, None]  # shape (3, N)

    cov_matrix = np.matmul(centered_target, np.transpose(centered_source)) / n_points

    if np.isnan(cov_matrix).any():
        print("nPoints:", n_points)
        print(source_hom.shape)
        print(target_hom.shape)
        raise RuntimeError("There are NANs in the input.")

    u, diag_values, vh = np.linalg.svd(cov_matrix, full_matrices=True)
    diag = np.diag(diag_values)
    s = np.eye(3)
    if np.linalg.det(cov_matrix) < 0.0:
        s[-1, -1] = -1

    rotation = u @ s @ vh

    var_p = np.var(source_hom[:3, :], axis=1).sum()
    if var_p == 0:
        print("Pose estimation failed: 0 variance in sampled points.""")
        print(source_hom)
        raise PoseEstimationError()
    scale_fact = 1 / var_p * np.trace(s @ diag)  # scale factor
    scales = np.array([scale_fact, scale_fact, scale_fact])
    scales_matrix = np.diag(scales)

    translation = target_centroid - scales_matrix @ rotation @ source_centroid

    # create homogeneous transformation from scale, rotation and translation
    out_transform = np.identity(4)
    out_transform[:3, :3] = scales_matrix @ rotation
    out_transform[:3, 3] = translation

    return scales, rotation, translation, out_transform


class PoseEstimationError(Exception):
    """Error if pose estimation encountered an error."""

    pass



"""Script to preprocess shapenet like dataset.

This script allows to interactively go through a ShapeNet dataset and decide
whether to keep or remove a mesh. All textures are removed, only the obj file
and a converted SDF volume are stored in the output folder."""
import argparse
import json
from multiprocessing import Pool, Process, Manager
import os
import shutil

from sdfest.vae import sdf_utils  # needs to be imported before pyrender

from typing import List
import numpy as np
import time
from tqdm import tqdm
import trimesh
from trimesh import Trimesh
from trimesh.scene.scene import Scene
from trimesh.visual.material import SimpleMaterial
import pyrender


class Object3D:
    """Representation of a 3D object."""

    def __init__(self, mesh_path: str):
        """Initialize the 3D object.

        Args:
            mesh_path: the full file path of the mesh
        """
        self.mesh_path = mesh_path
        self.simplified_mesh = None
        self.sdf_volume = None
        self.reconstructed_mesh = None

    def convert_to_sdf(self, cells_per_dim, padding):
        self.sdf_volume = sdf_utils.mesh_to_sdf(
            self.simplified_mesh, cells_per_dim, padding
        )
        return self

    def load_mesh(self):
        loaded_obj = trimesh.load(self.mesh_path, process=False)
        if isinstance(loaded_obj, Trimesh):
            self.simplified_mesh = Trimesh(
                loaded_obj.vertices,
                loaded_obj.faces,
                vertex_normals=loaded_obj.vertex_normals,
                visual=trimesh.visual.TextureVisuals(material=SimpleMaterial()),
            )
        elif isinstance(loaded_obj, Scene):
            mesh = trimesh.util.concatenate(
                tuple(
                    trimesh.Trimesh(
                        vertices=g.vertices,
                        faces=g.faces,
                        vertex_normals=g.vertex_normals,
                    )
                    for g in loaded_obj.geometry.values()
                )
            )
            self.simplified_mesh = Trimesh(
                mesh.vertices,
                mesh.faces,
                vertex_normals=mesh.vertex_normals,
                visual=trimesh.visual.TextureVisuals(material=SimpleMaterial()),
            )
        else:
            print(f"Not supported: {type(loaded_obj)}")
            return False
        return True

    def reconstruct_from_sdf(self):
        level = 1.0 / self.sdf_volume.shape[-1]
        self.reconstructed_mesh = sdf_utils.mesh_from_sdf(self.sdf_volume, level)
        return self


def trimesh_decision_viewer(mesh, return_dict):
    """View a trimesh with pyrender adding support to press left and right."""

    def on_press(key):
        nonlocal inp
        try:
            inp = key.char  # single-char keys
        except AttributeError:
            inp = key.name

    pyrender_mesh = pyrender.Mesh.from_trimesh(mesh)
    scene = pyrender.Scene(ambient_light=[0.3, 0.3, 0.3, 1.0])
    scene.add(pyrender_mesh)
    v = pyrender.Viewer(scene, run_in_thread=True, use_raymond_lighting=True)
    inp = None
    from pynput import keyboard

    decision = None

    listener = keyboard.Listener(on_press=on_press)
    listener.start()  # start to listen on a separate thread
    while True:
        time.sleep(0.1)
        if inp == "left":
            decision = "remove"
            break
        elif inp == "right":
            decision = "keep"
            break
        elif inp == "down":
            decision = "stop"
            break
    # TODO: listener.stop() hangs randomly (especially under load)
    v.close_external()
    return_dict["decision"] = decision
    return decision


def check_meshes(path):
    objects = []
    for root, _, files in os.walk(path):
        for file in files:
            _, file_extension = os.path.splitext(file)
            if file_extension == ".obj":
                full_file_path = os.path.join(root, file)
                objects.append(Object3D(full_file_path))

    print("Press left to discard the mesh, press right to keep it.")
    # first filtering
    print(
        "Decide which shapes generally fit your criteria? (irrespective of SDF "
        "conversion)"
    )
    total_objects = len(objects)
    filtered_objects = []
    for i, object in enumerate(objects):
        if object.load_mesh():
            manager = Manager()
            return_dict = manager.dict()
            p = Process(
                target=trimesh_decision_viewer,
                args=(object.simplified_mesh, return_dict),
            )
            p.start()
            p.join()
            decision = return_dict["decision"]
            if decision == "keep":
                filtered_objects.append(object)
            if decision == "stop":
                break
            progress = i / total_objects * 100
            print(f"Progress: {progress:.2f}% Kept: {len(filtered_objects)}")
    return filtered_objects


def conv_to_sdf(obj: Object3D):
    print(obj.mesh_path)
    return obj.convert_to_sdf(resolution, padding)


def convert_to_sdf(objects: List[Object3D]) -> List[Object3D]:
    # convert to SDF
    # TODO do not keep everything in memory, store to disk instead

    num_objects = len(objects)
    with Pool() as p:
        objects = list(
            tqdm(
                p.imap(conv_to_sdf, objects),
                total=len(objects),
            )
        )
    # only keep objects with valid SDF volume
    objects = [obj for obj in objects if obj.sdf_volume is not None]

    # convert to Mesh
    for obj in objects:
        obj.reconstruct_from_sdf()
    # only keep objects with valid mesh reconstruction
    objects = [obj for obj in objects if obj.reconstructed_mesh is not None]

    print(f"{len(objects)} / {num_objects} have been successfully converted.")

    return objects


if __name__ == "__main__":
    # define the arguments
    parser = argparse.ArgumentParser(description="Training script for init network.")

    # parse arguments int(float(x)) to support exponential notation
    parser.add_argument("--resolution", required=True, type=lambda x: int(x))
    parser.add_argument("--inpath", required=True)
    parser.add_argument("--outpath", required=True)
    parser.add_argument("--all", default=False, action="store_true")
    parser.add_argument("--padding", required=True, type=lambda x: int(x))
    parser.add_argument("--good_meshes", required=False, default="./good_meshes.json")
    parser.add_argument("--final_meshes", required=False, default="./final_meshes.json")

    args = parser.parse_args()

    path = args.inpath
    output_path = args.outpath
    resolution = args.resolution
    padding = args.padding
    no_filtering = args.all
    good_mesh_paths_file_path = args.good_meshes
    final_mesh_paths_file_path = args.final_meshes

    if not os.path.isfile(good_mesh_paths_file_path):
        good_mesh_paths_dict = {}
    else:
        with open(good_mesh_paths_file_path) as f:
            good_mesh_paths_dict = json.load(f)
    if not os.path.isfile(final_mesh_paths_file_path):
        final_mesh_paths_dict = {}
    else:
        with open(final_mesh_paths_file_path) as f:
            final_mesh_paths_dict = json.load(f)

    if no_filtering:
        objects = []
        for root, dirs, files in os.walk(path):
            for file in files:
                file_name, file_extension = os.path.splitext(file)
                if file_extension == ".obj":
                    full_file_path = os.path.join(root, file)
                    objects.append(full_file_path)
                    objects.append(Object3D(full_file_path))
                    objects[-1].load_mesh()
    else:
        if path in final_mesh_paths_dict:
            final_mesh_paths = final_mesh_paths_dict[path]
            objects = [Object3D(p) for p in final_mesh_paths]
            for object in objects:
                object.load_mesh()
        elif path in good_mesh_paths_dict:
            good_mesh_paths = good_mesh_paths_dict[path]
            objects = [Object3D(good_mesh_path) for good_mesh_path in good_mesh_paths]
            for object in objects:
                object.load_mesh()
        else:
            objects = check_meshes(path)
            while True:
                print(f"Store decisions? ({good_mesh_paths_file_path}) (y/n)")
                decision = input()
                if decision == "y":
                    good_mesh_paths = [obj.mesh_path for obj in objects]
                    good_mesh_paths_dict[path] = good_mesh_paths
                    with open(good_mesh_paths_file_path, "w") as f:
                        json.dump(good_mesh_paths_dict, f, indent=4)
                    break
                elif decision == "n":
                    break

    objects = convert_to_sdf(objects)

    # manually check if sdf volume is okay
    if no_filtering or path in final_mesh_paths_dict:
        final_objects = objects
    else:
        print("Decide which shapes have an okay sdf reconstruction?")
        final_objects = []
        for i, object in enumerate(objects):
            manager = Manager()
            return_dict = manager.dict()
            p = Process(
                target=trimesh_decision_viewer,
                args=(object.reconstructed_mesh, return_dict),
            )
            p.start()
            p.join()
            decision = return_dict["decision"]
            if decision == "keep":
                final_objects.append(object)
            if decision == "stop":
                break
            print(f"Progress: {i/len(objects) * 100:.2f}% Kept: {len(final_objects)}")
        while True:
            print(f"Store final decisions? ({final_mesh_paths_file_path}) (y/n)")
            decision = input()
            if decision == "y":
                final_mesh_paths = [obj.mesh_path for obj in final_objects]
                final_mesh_paths_dict[path] = final_mesh_paths
                with open(final_mesh_paths_file_path, "w") as f:
                    json.dump(final_mesh_paths_dict, f, indent=4)
                break
            elif decision == "n":
                break

    # save the objects
    os.makedirs(output_path, exist_ok=True)
    counter = 0
    for object in final_objects:
        output_file_path_obj = os.path.join(output_path, f"{counter:05}.obj")
        shutil.copyfile(object.mesh_path, output_file_path_obj)
        output_file_path_npy = os.path.join(output_path, f"{counter:05}.npy")
        np.save(output_file_path_npy, object.sdf_volume)
        counter += 1



"""Simple script to run inference on real data.

Usage (evaluation on random RGB-D images from folder):
    python -m sdfest.estimation.scripts.real_data \
        --config estimation/configs/rgbd_objects_uw.yaml estimation/configs/mug.yaml \
        --folder data/rgbd_objects_uw/coffee_mug/

Usage (evaluation on single RGB image from Redwood or RGB-D objects dataset):
    python -m sdfest.estimation.scripts.real_data \
        --config configs/rgbd_objects_uw.yaml configs/mug.yaml \
        --input rgbd_objects_uw/coffee_mug/coffee_mug_1/coffee_mug_1_1_103.png

Specific parameters:
    measure_runtime:
        if True, a breakdown of the runtime will be generated
        only supported for single input
    out_folder:
        if provided and measure_runtime is true, the runtime results are written to file
    visualize_optimization: whether to visualize optimization while at it
    visualize_input: whether to visualize the input
    create_animation:
        If true, three animations will be created. One for depth optimization, depth
        error, and mesh.
"""
import argparse
from collections import defaultdict
from datetime import datetime
import copy
import random
import os
import glob
import time
from typing import Dict, List, Callable, Tuple
import functools

import detectron2
import detectron2.engine
from detectron2 import model_zoo
from detectron2.utils.visualizer import Visualizer
from detectron2.data import MetadataCatalog
import numpy as np
import open3d as o3d
import pickle
import torch
import matplotlib.pyplot as plt
import yoco

import sdfest
from sdfest.estimation.simple_setup import SDFPipeline, NoDepthError


def load_real275_rgbd(rgb_path: str) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load RGB-D image from RGB path.

    Args:
        rgb_path: path to RGB image

    Returns:
        Tuple containing:
            - The color image, float32, RGB, 0-1, shape (H,W,C).
            - The depth image, float32, in meters, shape (H,W).
            - The color path.
            - The depth path.
    """
    depth_path = rgb_path[:-10] + "_depth.png"
    color_img = np.asarray(o3d.io.read_image(rgb_path), dtype=np.float32) / 255
    depth_img = (
        np.asarray(
            o3d.io.read_image(depth_path),
            dtype=np.float32,
        )
        * 0.001
    )
    return color_img, depth_img, rgb_path, depth_path


def load_real275_sample(folder: str) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load a sample from RGBD Object dataset.

    https://rgbd-dataset.cs.washington.edu/dataset/

    Args:
        folder: The root folder of the dataset.

    Returns:
        See load_real275_rgbd.
    """
    files = glob.glob(folder + "/**/*color.png", recursive=True)
    rgb_path = random.choice(files)
    return load_real275_rgbd(rgb_path)


def load_rgbd_object_uw_rgbd(rgb_path: str) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load RGB-D image from RGB path.

    Args:
        rgb_path: path to RGB image

    Returns:
        Tuple containing:
            - The color image, float32, RGB, 0-1, shape (H,W,C).
            - The depth image, float32, in meters, shape (H,W).
            - The color path.
            - The depth path.
    """
    depth_path = rgb_path[:-4] + "_depth" + rgb_path[-4:]
    color_img = np.asarray(o3d.io.read_image(rgb_path), dtype=np.float32) / 255
    depth_img = (
        np.asarray(
            o3d.io.read_image(depth_path),
            dtype=np.float32,
        )
        * 0.001
    )
    return color_img, depth_img, rgb_path, depth_path


def load_rgbd_object_uw_sample(folder: str) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load a sample from RGBD Object dataset.

    https://rgbd-dataset.cs.washington.edu/dataset/

    Args:
        folder: The root folder of the dataset.

    Returns:
        See load_rgbd_object_uw_rgbd.
    """
    files = glob.glob(folder + "/**/*[0-9].png", recursive=True)
    rgb_path = random.choice(files)
    return load_rgbd_object_uw_rgbd(rgb_path)


def load_redwood_rgbd(rgb_path: str) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load RGB-D image from RGB path of Redwood dataset.

    Args:
        rgb_path: path to RGB image

    Returns:
        Tuple containing:
            - The color image, float32, RGB, 0-1, shape (H,W,C).
            - The depth image, float32, in meters, shape (H,W).
            - The color path.
            - The depth path.
    """
    depth_dir = os.path.join(os.path.dirname(rgb_path), "..", "depth")

    rgb_timestamp = int(rgb_path[-16:-4])

    # find closest depth image in time
    depth_paths = glob.glob(depth_dir + "/*.png")
    depth_timestamps = np.array([int(p[-16:-4]) for p in depth_paths])
    ind = np.argmin(np.abs(depth_timestamps - rgb_timestamp))
    depth_path = depth_paths[ind]

    color_img = np.asarray(o3d.io.read_image(rgb_path), dtype=np.float32) / 255

    depth_img = (
        np.asarray(
            o3d.io.read_image(depth_path),
            dtype=np.float32,
        )
        * 0.001
    )
    return color_img, depth_img, rgb_path, depth_path


def load_redwood_sample(folder: str) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load a sample from Redwood dataset.

    Args:
        folder: The root folder of the dataset.

    Returns:
        Tuple containing:
            - The color image, float32, RGB, 0-1, shape (H,W,C).
            - The depth image, float32, in meters, shape (H,W).
            - The color path.
            - The depth path.
    """
    sequence_paths = glob.glob(folder + "/*")
    sequence_path = random.choice(sequence_paths)

    rgb_dir = os.path.join(sequence_path, "rgb")
    rgb_paths = glob.glob(rgb_dir + "/*.jpg")
    rgb_path = random.choice(rgb_paths)

    return load_redwood_rgbd(rgb_path)


def load_sample_from_folder(config: dict) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load a sample from dataset specified in config.

    See the dataset specific load functions for more details of the expected folder
    structure.

    Params:
        config: Configuration dictionary that must contain the following keys:
            "dataset": one of "redwood" | "rgbd_object_uw"
            "folder": the root folder of the dataset
    Returns:
        Tuple containing:
            - The color image, float32, RGB, 0-1, shape (H,W,C).
            - The depth image, float32, in meters, shape (H,W).
    """
    if config["dataset"] == "redwood":
        return load_redwood_sample(config["folder"])
    elif config["dataset"] == "rgbd_object_uw":
        return load_rgbd_object_uw_sample(config["folder"])
    elif config["dataset"] == "real275":
        return load_real275_sample(config["folder"])
    else:
        raise NotImplementedError(f"Dataset {config['dataset']} is not supported")


def _timing_decorator(timing_dict: defaultdict, name: str, f: Callable) -> Callable:
    @functools.wraps(f)
    def timing_wrapper(*args, **kwargs):
        before = time.time()
        result = f(*args, **kwargs)
        torch.cuda.synchronize()
        after = time.time()
        timing_dict[name].append([before, after])
        return result

    return timing_wrapper


def add_timing_decorators(pipeline: SDFPipeline) -> defaultdict:
    timing_dict = defaultdict(list)
    pipeline._nn_init = _timing_decorator(timing_dict, "init", pipeline._nn_init)
    pipeline.vae.decode = _timing_decorator(timing_dict, "decode", pipeline.vae.decode)
    pipeline.render = _timing_decorator(timing_dict, "render", pipeline.render)
    pipeline._compute_gradients = _timing_decorator(
        timing_dict, "backward", pipeline._compute_gradients
    )
    pipeline._compute_view_losses = _timing_decorator(
        timing_dict, "losses", pipeline._compute_view_losses
    )
    return timing_dict


def load_rgbd(config: dict) -> Tuple[np.ndarray, np.ndarray, str, str]:
    """Load a single RGB-D image from path and dataset specified in config.

    See the dataset specific load functions for more details of the expected folder
    structure.

    Params:
        config: Configuration dictionary that must contain the following keys:
            "dataset": one of "redwood" | "rgbd_object_uw"
            "input": the path to the RGB image
    Returns:
        Tuple containing:
            - The color image, float32, RGB, 0-1, shape (H,W,C).
            - The depth image, float32, in meters, shape (H,W).
            - The color path.
            - The depth path.
    """
    if config["dataset"] == "redwood":
        return load_redwood_rgbd(config["input"])
    elif config["dataset"] == "rgbd_object_uw":
        return load_rgbd_object_uw_rgbd(config["input"])
    elif config["dataset"] == "real275":
        return load_real275_rgbd(config["input"])
    else:
        raise NotImplementedError(f"Dataset {config['dataset']} is not supported")


def str_to_bool(v: str) -> bool:
    """Try to convert string to boolean.

    From: https://stackoverflow.com/a/43357954
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def generate_runtime_overview(config, timing_dicts: List[Dict]) -> None:
    total_runs = len(timing_dicts) / 2  # runs per setting, 2 settings
    stats_dicts = {
        True: defaultdict(lambda: {"total": 0.0, "total_calls": 0}),
        False: defaultdict(lambda: {"total": 0.0, "total_calls": 0}),
    }
    for timing_dict in timing_dicts:
        for name, timings in timing_dict.items():
            if name == "shape_optimization":
                continue
            for timing in timings:
                so = timing_dict["shape_optimization"]
                stats_dicts[so][name]["total"] += timing[1] - timing[0]
                stats_dicts[so][name]["total_calls"] += 1

    for stats_dict in stats_dicts.values():
        for name, stats in stats_dict.items():
            stats["mean"] = stats["total"] / stats["total_calls"]
            stats["calls_per_run"] = stats["total_calls"] / total_runs
            stats["total_per_run"] = stats["total"] / total_runs

    # save to file
    if "out_folder" in config and config["out_folder"] is not None:
        os.makedirs(config["out_folder"], exist_ok=True)
        filename = (
            f"runtime_analysis_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.yaml"
        )
        out_path = os.path.join(config["out_folder"], filename)
        combined_dict = {
            **config,
            "results_with_decode": dict(stats_dicts[True]),
            "results_without_decode": dict(stats_dicts[False]),
        }
        yoco.save_config_to_file(out_path, combined_dict)


def main() -> None:
    """Entry point of the program."""
    # define the arguments
    parser = argparse.ArgumentParser(description="SDF pose estimation in real data")

    # parse arguments
    parser.add_argument("--device")
    parser.add_argument("--input")
    parser.add_argument("--folder")
    parser.add_argument("--measure_runtime", type=str_to_bool, default=False)
    parser.add_argument("--visualize_optimization", type=str_to_bool, default=False)
    parser.add_argument("--visualize_input", type=str_to_bool, default=False)
    parser.add_argument("--cached_segmentation", action="store_true")
    parser.add_argument("--segmentation_dir", default="./cached_segmentations/")
    parser.add_argument("--config", default="configs/default.yaml", nargs="+")

    config = yoco.load_config_from_args(
        parser, search_paths=[".", "~/.sdfest/", sdfest.__path__[0]]
    )

    if "input" in config and "folder" in config:
        print("Only one of input and folder can be specified.")
        exit()

    if config["measure_runtime"] and config["visualize_optimization"]:
        print("Visualization not supported while measuring runtime.")
        exit()

    pipeline = SDFPipeline(config)
    create_animation = (
        config["create_animation"] if "create_animation" in config else False
    )

    timing_dict = None
    timing_dicts = []
    if config["measure_runtime"]:
        timing_dict = add_timing_decorators(pipeline)

    # Segmentation using detectron2
    print("Loading segmentation model...")
    cfg = detectron2.config.get_cfg()
    cfg.merge_from_file(
        model_zoo.get_config_file(
            "COCO-InstanceSegmentation/mask_rcnn_X_101_32x8d_FPN_3x.yaml"
        )
    )
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5  # set threshold for this model
    cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(
        "COCO-InstanceSegmentation/mask_rcnn_X_101_32x8d_FPN_3x.yaml"
    )

    predictor = detectron2.engine.DefaultPredictor(cfg)
    print("Segmentation model loaded.")

    completed_runs = 0
    shape_optimization = True

    while True:
        if timing_dict is not None:
            timing_dict.clear()
        if "folder" in config:
            color_img, depth_img, color_path, _ = load_sample_from_folder(config)
        elif "input" in config:
            color_img, depth_img, color_path, _ = load_rgbd(config)
        else:
            print("No folder or input file specified.")
            exit()

        if timing_dict is not None:
            timing_dict["pipeline"].append([time.time(), None])
            timing_dict["segmentation"].append([time.time(), None])

        if config["cached_segmentation"]:
            # check if segmentation exists
            color_name, _ = os.path.splitext(color_path)
            color_dir = os.path.dirname(color_name)
            segmentation_dir = os.path.join(config["segmentation_dir"], color_dir)
            segmentation_path = (
                os.path.join(config["segmentation_dir"], color_name) + ".pickle"
            )
            os.makedirs(segmentation_dir, exist_ok=True)

            if os.path.isfile(segmentation_path):
                with open(segmentation_path, "rb") as f:
                    outputs = pickle.load(f)
            else:
                # compute segmentation and save
                # detectron expects (H,C,W), BGR, 0-255 as input
                detectron_color_img = color_img[:, :, ::-1] * 255
                outputs = predictor(detectron_color_img)
                with open(segmentation_path, "wb") as f:
                    pickle.dump(outputs, f)
        else:
            # detectron expects (H,C,W), BGR, 0-255 as input
            detectron_color_img = color_img[:, :, ::-1] * 255
            outputs = predictor(detectron_color_img)

        if timing_dict is not None:
            torch.cuda.synchronize()
            timing_dict["segmentation"][-1][1] = time.time()

        category_id = MetadataCatalog.get(cfg.DATASETS.TRAIN[0]).thing_classes.index(
            config["category"]
        )
        matching_instances = []
        for i in range(len(outputs["instances"])):
            instance = outputs["instances"][i]
            if instance.pred_classes != category_id:
                continue
            matching_instances.append(instance)

        matching_instances.sort(key=lambda k: k.pred_masks.sum())

        if not matching_instances:
            print("Warning: category not detected in input")
        else:
            print("Category detected")

        for instance in matching_instances:
            if create_animation:
                animation_path = os.path.join(
                    os.getcwd(),
                    f"animation_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}",
                )
            else:
                animation_path = None

            if config["visualize_input"]:
                v = Visualizer(
                    color_img * 255,
                    MetadataCatalog.get(cfg.DATASETS.TRAIN[0]),
                    scale=1.2,
                )
                out = v.draw_instance_predictions(instance.to("cpu"))
                plt.imshow(out.get_image())
                plt.show()
                plt.imshow(depth_img)
                plt.show()
            depth_tensor = torch.from_numpy(depth_img).to(config["device"])
            instance_mask = instance.pred_masks.cuda()[0]
            color_img_tensor = torch.from_numpy(color_img).to(config["device"])

            try:
                position, orientation, scale, shape = pipeline(
                    depth_tensor,
                    instance_mask,
                    color_img_tensor,
                    visualize=config["visualize_optimization"],
                    animation_path=animation_path,
                    shape_optimization=shape_optimization,
                )
            except NoDepthError:
                print("No depth data, skipping")

            break  # comment to evaluate all instances, instead of largest only

        if timing_dict is not None:
            torch.cuda.synchronize()
            timing_dict["pipeline"][-1][1] = time.time()

            if completed_runs != 0 or not config["skip_first_run"]:
                timing_dicts.append(copy.deepcopy(timing_dict))
                timing_dicts[-1]["shape_optimization"] = shape_optimization

            print(f"\r{completed_runs+1}/{config['runs']}", end="")

        completed_runs += 1

        # only run single evaluation for single file
        if "input" in config and not config["measure_runtime"]:
            break
        elif config["measure_runtime"] and completed_runs == config["runs"]:
            if shape_optimization:
                shape_optimization = False
                completed_runs = 0
                print("")
            else:
                print("")
                break

    if config["measure_runtime"]:
        generate_runtime_overview(config, timing_dicts)


if __name__ == "__main__":
    main()



"""Module which provides SDFDataset class."""
import os
import torch
import numpy as np


class SDFDataset(torch.utils.data.Dataset):
    """Dataset of SDF volumes stored in .npy format.

    Expected dataset format:
        {sdf_folder}/00000.npy
        {sdf_folder}/00001.npy
        ...
    """

    def __init__(self, sdf_folder: str):
        """Construct the dataset.

        Args:
            sdf_folder: The folder containing the npy files.
        """
        self.path = sdf_folder

        self.size = len([f for f in os.listdir(sdf_folder) if f.endswith(".npy")])

    def __len__(self):
        """Return the number of images in the dataset."""
        return self.size

    def __getitem__(self, idx: int) -> torch.Tensor:
        """Return SDF volume at a specific index.

        Args:
            idx: The index of the sdf file to retrieve.
        Returns:
            The loaded SDF volume.
        """
        sdf_path = os.path.join(self.path, f"{idx:05}.npy")
        sdf_np = np.load(sdf_path)
        return torch.as_tensor(sdf_np).unsqueeze(0)



"""Parametrized networks for pose and shape estimation."""
import torch
import torch.nn as nn
from typing import List, Callable, Optional

from sdfest.initialization.so3grid import SO3Grid


class SDFPoseHead(nn.Module):
    """Parametrized head to estimate pose and shape from feature vector."""

    def __init__(
        self,
        in_size: int,
        mlp_out_sizes: List,
        shape_dimension: int,
        batchnorm: bool,
        orientation_repr: Optional[str] = "quaternion",
        orientation_grid_resolution: Optional[int] = None,
    ):
        """Initialize the SDFPoseHead.

        Args:
            in_size:            number of input features
            mlp_out_sizes:      output sizes of each linear layer
            shape_dimension:    dimension of shape description
            batchnorm:          whether to use batchnorm or not
            orientation_repr:
                The orientation represention. One of "quaternion"|"discretized".
            orientation_grid_resolution:
                The resolution of the SO3 grid.
                Only used when orientation_repr == "discretized".
        """
        super().__init__()

        self._in_size = in_size
        self._mlp_out_sizes = mlp_out_sizes
        self._batchnorm = batchnorm
        self._shape_dimension = shape_dimension
        self._orientation_repr = orientation_repr

        # define layers
        self._linear_layers = torch.nn.ModuleList([])
        for i, out_size in enumerate(mlp_out_sizes):
            if i == 0:
                self._linear_layers.append(nn.Linear(self._in_size, out_size))
            else:
                self._linear_layers.append(nn.Linear(mlp_out_sizes[i - 1], out_size))

        self._bn_layers = torch.nn.ModuleList([])
        if self._batchnorm:
            for out_size in mlp_out_sizes:
                self._bn_layers.append(nn.BatchNorm1d(out_size))

        if orientation_repr == "quaternion":
            self._grid = None
            self._final_layer = nn.Linear(mlp_out_sizes[-1], self._shape_dimension + 8)
        elif orientation_repr == "discretized":
            self._grid = SO3Grid(orientation_grid_resolution)
            self._final_layer = nn.Linear(
                mlp_out_sizes[-1], self._shape_dimension + 4 + self._grid.num_cells()
            )
        else:
            raise NotImplementedError(
                f"orientation_repr {orientation_repr} is not supported."
            )

    def forward(self, x):
        """Forward pass of the module.

        Input represents set of input features used to compute pose.

        Args:
            x: batch of input vectors
        Returns:
            Tuple with the following entries:
                The predicted shape vector.
                The predicted pose.
                The predicted scale.
                The predicted orientation in the specified orientation representation.
                    For "quaternion" this will be of shape (N,4) with each quaternion
                    having the order (x, y, z, w), i.e., scalar-last, and normalized.
                    For "discretized" this will be of shape (N,M) based on the grid
                    resolution. No activation function is applied. I.e., softmax has
                    to be used to get probabilities, and cross_entropy_loss should be
                    used during training.
        """
        out = x
        for i, linear_layer in enumerate(self._linear_layers):
            out = linear_layer(out)
            if self._batchnorm:
                out = self._bn_layers[i](out)
            out = nn.functional.relu(out)

        # Normalize quaternion
        if self._orientation_repr == "quaternion":
            out = self._final_layer(out)
            orientation = out[:, self._shape_dimension + 4 :]
            orientation = orientation / torch.sqrt(
                torch.sum(orientation ** 2, 1, keepdim=True)
            )
        elif self._orientation_repr == "discretized":
            out = self._final_layer(out)
            orientation = out[:, self._shape_dimension + 4 :]
        else:
            raise NotImplementedError(
                f"orientation_repr {self.orientation_repr} is not supported."
            )

        return (
            out[:, 0 : self._shape_dimension],
            out[:, self._shape_dimension : self._shape_dimension + 3],
            out[:, self._shape_dimension + 3],
            orientation,
        )


class SDFPoseNet(nn.Module):
    """Pose and shape estimation from sensor data.

    Composed of feature extraction backbone and shape/pose head.
    """

    def __init__(self, backbone: nn.Module, head: nn.Module):
        """Construct SDF pose and shape network.

        Args:
            backbone:       function or class representing the backbone
            backbone_dict:  parameters passed to backbone on construction
            head:           function or class representing the head
            head_dict:      parameters passed to head on construction
        """
        super().__init__()
        self._backbone = backbone
        self._head = head

    def forward(self, x):
        """Forward pass.

        Args:
            x: input compatible with backbone.
        Returns:
            output from head
        """
        out = self._backbone(x)
        out = self._head(out)
        return out


if __name__ == "__main__":
    # simple sanity check
    sdf_pose_head = SDFPoseHead(
        1024,
        [512, 256, 128],
        10,
        True,
        orientation_repr="discretized",
        orientation_grid_resolution=0,
    )
    inp = torch.rand(16, 1024)  # batch size 16 with 1024 features
    out = sdf_pose_head(inp)
    print([x.shape for x in out])
